# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
# (C) Copyright 2007 Riverbank Computing Limited
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


import sys

from traits.api import observe, provides, Str

from pyface.qt import QtGui
from pyface.i_application_window import IApplicationWindow, MApplicationWindow
from .image_resource import ImageResource
from .window import Window


@provides(IApplicationWindow)
class ApplicationWindow(MApplicationWindow, Window):
    """ The toolkit specific implementation of an ApplicationWindow.  See the
    IApplicationWindow interface for the API documentation.
    """

    # 'IWindow' interface -------------------------------------------------#

    #: The window title.
    title = Str("Pyface")

    # ------------------------------------------------------------------------
    # Protected 'IApplicationWindow' interface.
    # ------------------------------------------------------------------------

    def _create_contents(self, parent):
        panel = QtGui.QWidget(parent)

        palette = QtGui.QPalette(panel.palette())
        palette.setColor(QtGui.QPalette.ColorRole.Window, QtGui.QColor("blue"))
        panel.setPalette(palette)
        panel.setAutoFillBackground(True)

        return panel

    def _create_menu_bar(self, parent):
        if self.menu_bar_manager is not None:
            menu_bar = self.menu_bar_manager.create_menu_bar(parent)
            self.control.setMenuBar(menu_bar)

    def _create_status_bar(self, parent):
        if self.status_bar_manager is not None:
            status_bar = self.status_bar_manager.create_status_bar(parent)

            # QMainWindow automatically makes the status bar visible, but we
            # have delegated this responsibility to the status bar manager.
            self.control.setStatusBar(status_bar)
            status_bar.setVisible(self.status_bar_manager.visible)

    def _create_tool_bar(self, parent):
        tool_bar_managers = self.tool_bar_managers
        visible = self.control.isVisible()
        for tool_bar_manager in tool_bar_managers:
            # Add the tool bar and make sure it is visible.
            tool_bar = tool_bar_manager.create_tool_bar(parent)
            self.control.addToolBar(tool_bar)
            tool_bar.show()

            # Make sure that the tool bar has a name so its state can be saved.
            if len(tool_bar.objectName()) == 0:
                tool_bar.setObjectName(tool_bar_manager.name)

        if sys.platform == "darwin":
            # Work around bug in Qt on OS X where creating a tool bar with a
            # QMainWindow parent hides the window. See
            # http://bugreports.qt.nokia.com/browse/QTBUG-5069 for more info.
            self.control.setVisible(visible)

    def _set_window_icon(self):
        if self.icon is None:
            icon = ImageResource("application.png")
        else:
            icon = self.icon
        if self.control is not None:
            self.control.setWindowIcon(icon.create_icon())

    # ------------------------------------------------------------------------
    # 'Window' interface.
    # ------------------------------------------------------------------------

    def _size_default(self):
        """ Trait initialiser. """

        return (800, 600)

    # ------------------------------------------------------------------------
    # Protected 'IWidget' interface.
    # ------------------------------------------------------------------------

    def create(self, parent=None):
        super().create(parent=parent)

        contents = self._create_contents(self.control)
        self.control.setCentralWidget(contents)

        self._create_trim_widgets(self.control)

    def _create_control(self, parent):
        control = super()._create_control(parent)
        control.setObjectName("ApplicationWindow")

        control.setAnimated(False)
        control.setDockNestingEnabled(True)

        return control

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    # Trait change handlers ------------------------------------------------

    # QMainWindow takes ownership of the menu bar and the status bar upon
    # assignment. For this reason, it is unnecessary to delete the old controls
    # in the following two handlers.

    @observe("menu_bar_manager")
    def _menu_bar_manager_updated(self, event):
        if self.control is not None:
            self._create_menu_bar(self.control)

    @observe("status_bar_manager")
    def _status_bar_manager_updated(self, event):
        if self.control is not None:
            if event.old is not None:
                self.control.setStatusBar(None)
                event.old.destroy()
            self._create_status_bar(self.control)

    @observe("tool_bar_managers.items")
    def _update_tool_bar_managers(self, event):
        if self.control is not None:
            # Remove the old toolbars.
            for child in self.control.children():
                if isinstance(child, QtGui.QToolBar):
                    self.control.removeToolBar(child)
                    child.deleteLater()

            # Add the new toolbars.
            if event.new is not None:
                self._create_status_bar(self.control)

    @observe("icon")
    def _icon_updated(self, event):
        self._set_window_icon()



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Heading text. """

import warnings

from traits.api import HasTraits, Int, Interface, Str


class IHeadingText(Interface):
    """ A widget which shows heading text in a panel. """

    # 'IHeadingText' interface ---------------------------------------------

    #: Heading level.  This is currently unused.
    level = Int(1)

    #: The heading text.
    text = Str("Default")


class MHeadingText(HasTraits):
    """ The mixin class that contains common code for toolkit specific
    implementations of the IHeadingText interface.
    """

    # 'IHeadingText' interface ---------------------------------------------

    #: Heading level.  This is currently unused.
    level = Int(1)

    #: The heading text.
    text = Str("Default")

    def __init__(self, parent=None, **traits):
        """ Creates the heading text. """

        if "image" in traits:
            warnings.warn(
                "background images are no-longer supported for Wx and the "
                "'image' trait will be removed in a future Pyface update",
                DeprecationWarning,
                stacklevel=2,
            )

        create = traits.pop("create", None)

        # Base class constructor.
        super().__init__(parent=parent, **traits)

        if create:
            # Create the widget's toolkit-specific control.
            self.create()
            warnings.warn(
                "automatic widget creation is deprecated and will be removed "
                "in a future Pyface version, code should not pass the create "
                "parameter and should instead call create() explicitly",
                DeprecationWarning,
                stacklevel=2,
            )
        elif create is not None:
            warnings.warn(
                "setting create=False is no longer required",
                DeprecationWarning,
                stacklevel=2,
            )

    # ------------------------------------------------------------------------
    # 'IWidget' interface.
    # ------------------------------------------------------------------------

    def _initialize_control(self):
        """ Perform any toolkit-specific initialization for the control. """
        super()._initialize_control()
        self._set_control_text(self.text)

    def _add_event_listeners(self):
        super()._add_event_listeners()
        self.observe(self._text_updated, 'text', dispatch="ui")

    def _remove_event_listeners(self):
        self.observe(self._text_updated, 'text', dispatch="ui", remove=True)
        super()._remove_event_listeners()

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _set_control_text(self, text):
        """ Set the text on the control.

        Parameters
        ----------
        text : str
            The text to display.  This can contain basic HTML-like markeup.
        """
        raise NotImplementedError()

    def _get_control_text(self):
        """ Get the text on the control.

        Returns
        ----------
        text : str
            The text to displayed in the widget.
        """
        raise NotImplementedError()

    # Trait change handlers --------------------------------------------------

    def _text_updated(self, event):
        if self.control is not None:
            self._set_control_text(self.text)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A dialog that allows the user to select a font. """

from pyface.qt import QtGui

from traits.api import provides

from pyface.font import Font
from pyface.ui_traits import PyfaceFont
from pyface.i_font_dialog import IFontDialog
from .dialog import Dialog


@provides(IFontDialog)
class FontDialog(Dialog):
    """ A dialog that allows the user to choose a font.
    """

    # 'IFontDialog' interface ----------------------------------------------

    #: The font in the dialog.
    font = PyfaceFont()

    # ------------------------------------------------------------------------
    # 'IDialog' interface.
    # ------------------------------------------------------------------------

    def _create_contents(self, parent):
        # In PyQt this is a canned dialog so there are no contents.
        pass

    # ------------------------------------------------------------------------
    # 'IWindow' interface.
    # ------------------------------------------------------------------------

    def close(self):
        if self.control.result() == QtGui.QDialog.DialogCode.Accepted:
            qfont = self.control.selectedFont()
            self.font = Font.from_toolkit(qfont)
        return super(FontDialog, self).close()

    # ------------------------------------------------------------------------
    # 'IWindow' interface.
    # ------------------------------------------------------------------------

    def _create_control(self, parent):
        qfont = self.font.to_toolkit()
        dialog = QtGui.QFontDialog(qfont, parent)
        return dialog



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The field interface. """


from traits.api import Any, HasTraits

from pyface.i_layout_widget import ILayoutWidget
from pyface.ui_traits import Alignment


class IField(ILayoutWidget):
    """ The field interface.

    A field is a widget that displays a value and (potentially) allows a user
    to interact with it.
    """

    #: The value held by the field.
    value = Any()

    #: The alignment of the field's content.
    alignment = Alignment()


class MField(HasTraits):
    """ The field mix-in. """

    #: The value held by the field.
    value = Any()

    #: The alignment of the text in the field.
    alignment = Alignment()

    # ------------------------------------------------------------------------
    # IWidget interface
    # ------------------------------------------------------------------------

    def create(self, parent=None):
        """ Creates the toolkit specific control.

        This method should create the control and assign it to the
        :py:attr:``control`` trait.
        """
        super().create(parent=parent)

        self.show(self.visible)
        self.enable(self.enabled)

    def _initialize_control(self):
        """ Perform any post-creation initialization for the control.
        """
        super()._initialize_control()
        self._set_control_value(self.value)
        if self.alignment != 'default':
            self._set_control_alignment(self.alignment)

    def _add_event_listeners(self):
        """ Set up toolkit-specific bindings for events """
        super()._add_event_listeners()
        self.observe(self._value_updated, "value", dispatch="ui")
        self.observe(self._alignment_updated, "alignment", dispatch="ui")

    def _remove_event_listeners(self):
        """ Remove toolkit-specific bindings for events """
        self.observe(
            self._value_updated, "value", dispatch="ui", remove=True
        )
        self.observe(
            self._alignment_updated, "alignment", dispatch="ui", remove=True
        )
        super()._remove_event_listeners()

    # ------------------------------------------------------------------------
    # Private interface
    # ------------------------------------------------------------------------

    def _get_control(self):
        """ If control is not passed directly, get it from the trait. """
        control = self.control
        if control is None:
            raise RuntimeError("Toolkit control does not exist.")
        return control

    # Toolkit control interface ---------------------------------------------

    def _get_control_value(self):
        """ Toolkit specific method to get the control's value. """
        raise NotImplementedError()

    def _set_control_value(self, value):
        """ Toolkit specific method to set the control's value. """
        raise NotImplementedError()

    def _get_control_alignment(self):
        """ Toolkit specific method to get the control's read_only state. """
        raise NotImplementedError()

    def _set_control_alignment(self, alignment):
        """ Toolkit specific method to set the control's alignment. """
        raise NotImplementedError()

    # Trait change handlers -------------------------------------------------

    def _value_updated(self, event):
        value = event.new
        if self.control is not None:
            self._set_control_value(value)

    def _alignment_updated(self, event):
        alignment = event.new
        if self.control is not None:
            self._set_control_alignment(alignment)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Toolkit-independent font utilities.

Pyface fonts are intended to be generic, but able to be mapped fairly well
to most backend toolkit font descriptions.  In most cases we can describe
fonts along the common dimensions that are used by CSS, Wx, and Qt.  However
when it comes to actually working with a font, the toolkit needs to take the
description and produce something that is as close as possible to the
specification, but within the constraints of the toolkit, operating system
and available fonts on the machine where this is being executed.

Because of this inherent ambiguity in font specification, this system tries to
be flexible in what it accepts as a font specification, rather than trying to
specify a unique canoncial form.

Font Properties
---------------

The properties that fonts have are:

Font Family
    A list of font family names in order of preference, such as "Helvetica"
    or "Comic Sans".  In the case of a font that has been selected by the
    toolkit this list will have one value which is the actual font family name.

    There are several generic font family names that can be used as fall-backs
    in case all preferred fonts are unavailable.  The allowed values are:

    "default"
        The application's default system font.

    "fantasy"
        A primarily decorative font, but with recognisable characters.

    "decorative"
        A synonym for "fantasy".

    "serif"
        A proportional serif font, such as Times New Roman or Garamond.

    "roman"
        A synonym for "serif".

    "cursive"
        A font which resembles hand-written cursive text, such as Zapf
        Chancery.

    "script"
        A synonym for "cursive".

    "sans-serif"
        A proportional sans-serif font, such as Helvetica or Arial.

    "swiss"
        A synonym for "sans-serif".

    "monospace"
        A fixed-pitch sans-serif font, such as Source Code Pro or Roboto Mono.
        Commonly used for display of code.

    "modern"
        A synonym for "monospace".

    "typewriter"
        A fixed-pitch serif font which resembles typewritten text, such as
        Courier.  Commonly used for display of code.

    "teletype"
        A synonym for "typewriter".

    These special names will be converted into appropriate toolkit flags which
    correspond to these generic font specifications.

Weight
    How thick or dark the font glyphs are.  These can be given as a number
    from 1 (lightest) to 999 (darkest), but are typically specified by a
    multiple of 100 from 100 to 900, with a number of synonyms such as 'light'
    and 'bold' available for those values.

Stretch
    The amount of horizontal compression or expansion to apply to the glyphs.
    These can be given as a percentage between 50% and 200%, or by strings
    such as 'condensed' and 'expanded' that correspond to those values.

Style
    This selects either 'oblique' or 'italic' variants typefaces of the given
    font family.  If neither is wanted, the value is 'normal'.

Size
    The overall size of the glyphs. This can be expressed either as the
    numeric size in points, or as a string such as "small" or "large".

Variants
    A set of additional font style specifiers, such as "small-caps",
    "strikethrough", "underline" or "overline", where supported by the
    underlying toolkit.

Font Specificiation Class
-------------------------

The Pyface Font class is a HasStrictTraits class which specifies a requested
font.  It has methods that convert the Font class to and from a toolkit Font
class.

"""
from traits.api import (
    BaseCFloat, CList, CSet, Enum, HasStrictTraits, Map, Str,
)
from traits.trait_type import NoDefaultSpecified


#: Font weight synonyms.
#: These are alternate convenience names for font weights.
#: The intent is to allow a developer to use a common name (eg. "bold") instead
#: of having to remember the corresponding number (eg. 700).
#: These come from:
#: - the OpenType specification: https://docs.microsoft.com/en-us/typography/opentype/spec/os2#usweightclass
#: - QFont weights: https://doc.qt.io/qt-5/qfont.html#Weight-enum
#: - WxPython font weights: https://wxpython.org/Phoenix/docs/html/wx.FontWeight.enumeration.html
#: - CSS Common weight name mapping: https://developer.mozilla.org/en-US/docs/Web/CSS/font-weight#common_weight_name_mapping
#: - values used by Enable: https://github.com/enthought/enable/blob/78d2e494097fac71cc5c73efef5fb464963fb4db/kiva/fonttools/_constants.py#L90-L105
#: See also: https://gist.github.com/lukaszgrolik/5849599
WEIGHTS = {str(i): i for i in range(100, 1001, 100)}
WEIGHTS.update({
    'thin': 100,
    'hairline': 100,
    'extra-light': 200,
    'ultra-light': 200,
    'ultralight': 200,
    'light': 300,
    'normal': 400,
    'regular': 400,
    'book': 400,
    'medium': 500,
    'roman': 500,
    'semi-bold': 600,
    'demi-bold': 600,
    'demi': 600,
    'bold': 700,
    'extra-bold': 800,
    'ultra-bold': 800,
    'extra bold': 800,
    'black': 900,
    'heavy': 900,
    'extra-heavy': 1000,
})

#: Font stretch synonyms.
#: These are alternate convenience names for font stretch/width values.
#: The intent is to allow a developer to use a common name (eg. "expanded")
#: instead of having to remember the corresponding number (eg. 125).
#: These come from:
#: - the OpenType specification: https://docs.microsoft.com/en-us/typography/opentype/spec/os2#uswidthclass
#: - QFont stetch: https://doc.qt.io/qt-5/qfont.html#Stretch-enum
#: - CSS font-stretch: https://developer.mozilla.org/en-US/docs/Web/CSS/font-stretch
#: - values used by Enable: https://github.com/enthought/enable/blob/78d2e494097fac71cc5c73efef5fb464963fb4db/kiva/fonttools/_constants.py#L78-L88
STRETCHES = {
    'ultra-condensed': 50,
    'extra-condensed': 62.5,
    'condensed': 75,
    'semi-condensed': 87.5,
    'normal': 100,
    'semi-expanded': 112.5,
    'expanded': 125,
    'extra-expanded': 150,
    'ultra-expanded': 200,
}

#: Font size synonyms.
#: These are alternate convenience names for font size values.
#: The intent is to allow a developer to use a common name (eg. "small")
#: instead of having to remember the corresponding number (eg. 10).
#: These come from CSS font-size: https://developer.mozilla.org/en-US/docs/Web/CSS/font-size
SIZES = {
    'xx-small': 7.0,
    'x-small': 9.0,
    'small': 10.0,
    'medium': 12.0,
    'large': 14.0,
    'x-large': 18.0,
    'xx-large': 20.0,
}

STYLES = ('normal', 'italic', 'oblique')

#: Font variants. Currently only small caps variants are exposed in Qt, and
#: nothing in Wx.  In the future this could include things like swashes,
#: numeric variants, and so on, as exposed in the toolkit.
VARIANTS = ['small-caps']

#: Additional markings on or around the glyphs of the font that are not part
#: of the glyphs themselves.  Currently Qt and Wx support underline and
#: strikethrough, and Qt supports overline.  In the future overlines and other
#: decorations may be supported, as exposed in the toolkit.
DECORATIONS = ['underline', 'strikethrough', 'overline']

#: A trait for font families.
FontFamily = CList(Str, ['default'])

#: A trait for font weights.
FontWeight = Map(WEIGHTS, default_value='normal')

#: A trait for font styles.
FontStyle = Enum(STYLES)

#: A trait for font variant properties.
FontVariants = CSet(Enum(VARIANTS))

#: A trait for font decorator properties.
FontDecorations = CSet(Enum(DECORATIONS))


class FontStretch(BaseCFloat):
    """ Trait type for font stretches.

    The is a CFloat trait which holds floating point values between 50 and 200,
    inclusive.  In addition to values which can be converted to floats, this
    trait also accepts named synonyms for sizes which are converted to the
    associated commonly accepted weights:

    - 'ultra-condensed': 50
    - 'extra-condensed': 62.5
    - 'condensed': 75
    - 'semi-condensed': 87.5
    - 'normal': 100
    - 'semi-expanded': 112.5
    - 'expanded': 125
    - 'extra-expanded': 150
    - 'ultra-expanded': 200
    """

    #: The default value for the trait.
    default_value = 100.0

    def __init__(self, default_value=NoDefaultSpecified, **metadata):
        if default_value != NoDefaultSpecified:
            default_value = self.validate(None, None, default_value)
        super().__init__(default_value, **metadata)

    def validate(self, object, name, value):
        """Validate the trait

        This is a CFloat trait that also accepts percentage strings.  Values
        must be in the range 50 to 200 inclusive
        """
        if isinstance(value, str) and value.endswith('%'):
            value = value[:-1]
        value = STRETCHES.get(value, value)
        value = super().validate(object, name, value)
        if not 50 <= value <= 200:
            self.error(object, name, value)
        return value

    def info(self):
        """Describe the trait"""
        info = (
            "a float from 50 to 200, "
            "a value that can convert to a float from 50 to 200, "
        )
        info += ', '.join(repr(key) for key in SIZES)
        info += (
            " or a string with a float value from 50 to 200 followed by '%'"
        )
        return info


class FontSize(BaseCFloat):
    """ Trait type for font sizes.

    The is a CFloat trait which also allows values which are keys of the
    size dictionary, and also ignores trailing 'pt' ot 'px' annotation in
    string values.  The value stored is a float.
    """

    #: The default value for the trait.
    default_value = 12.0

    def __init__(self, default_value=NoDefaultSpecified, **metadata):
        if default_value != NoDefaultSpecified:
            default_value = self.validate(None, None, default_value)
        super().__init__(default_value, **metadata)

    def validate(self, object, name, value):
        """Validate the trait

        This is a CFloat trait that also accepts strings with 'pt' or 'px'
        suffixes.  Values must be positive.
        """
        if (
            isinstance(value, str)
            and (value.endswith('pt') or value.endswith('px'))
        ):
            value = value[:-2]
        value = SIZES.get(value, value)
        value = super().validate(object, name, value)
        if value <= 0:
            self.error(object, name, value)
        return value

    def info(self):
        """Describe the trait"""
        info = (
            "a positive float, a value that can convert to a positive float, "
        )
        info += ', '.join(repr(key) for key in SIZES)
        info += (
            " or a string with a positive float value followed by 'pt' or 'px'"
        )
        return info


class Font(HasStrictTraits):
    """A toolkit-independent font specification.

    This class represents a *request* for a font with certain characteristics,
    not a concrete font that can be used for drawing.  Font objects returned
    from the toolkit may or may not match what was requested, depending on the
    capabilities of the toolkit, OS, and the fonts installed on a particular
    computer.
    """

    #: The preferred font families.
    family = FontFamily()

    #: The weight of the font.
    weight = FontWeight()

    #: How much the font is expanded or compressed.
    stretch = FontStretch()

    #: The style of the font.
    style = FontStyle()

    #: The size of the font.
    size = FontSize()

    #: The font variants.
    variants = FontVariants()

    #: The font decorations.
    decorations = FontDecorations()

    @classmethod
    def from_toolkit(cls, toolkit_font):
        """ Create a Font from a toolkit font object.

        Parameters
        ----------
        toolkit_font : Any
            A toolkit font to be converted to a corresponding class instance,
            within the limitations of the options supported by the class.
        """
        from pyface.toolkit import toolkit_object
        toolkit_font_to_properties = toolkit_object(
            'font:toolkit_font_to_properties')

        return cls(**toolkit_font_to_properties(toolkit_font))

    def to_toolkit(self):
        """ Create a toolkit font object from the Font instance.

        Returns
        -------
        toolkit_font : Any
            A toolkit font which matches the property of the font as
            closely as possible given the constraints of the toolkit.
        """
        from pyface.toolkit import toolkit_object
        font_to_toolkit_font = toolkit_object('font:font_to_toolkit_font')

        return font_to_toolkit_font(self)

    def __str__(self):
        """ Produce a CSS2-style representation of the font. """
        terms = []
        if self.style != 'normal':
            terms.append(self.style)
        terms.extend(
            variant for variant in VARIANTS
            if variant in self.variants
        )
        terms.extend(
            decoration for decoration in DECORATIONS
            if decoration in self.decorations
        )
        if self.weight != 'normal':
            terms.append(self.weight)
        if self.stretch != 100:
            terms.append("{:g}%".format(self.stretch))
        size = self.size
        # if size is an integer we want "12pt" not "12.pt"
        if int(size) == size:
            size = int(size)
        terms.append("{}pt".format(size))
        terms.append(
            ', '.join(
                repr(family) if ' ' in family else family
                for family in self.family
            )
        )
        return ' '.join(terms)

    def __repr__(self):
        traits = self.trait_get(self.editable_traits())
        trait_args = ', '.join(
            "{}={!r}".format(name, value)
            for name, value in traits.items()
        )
        return "{}({})".format(self.__class__.__name__, trait_args)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


""" Enthought pyface package component
"""

import wx

from traits.api import Str, observe, provides

from pyface.i_application_window import IApplicationWindow, MApplicationWindow
from pyface.wx.aui import aui, PyfaceAuiManager
from .image_resource import ImageResource
from .window import Window


@provides(IApplicationWindow)
class ApplicationWindow(MApplicationWindow, Window):
    """ The toolkit specific implementation of an ApplicationWindow.  See the
    IApplicationWindow interface for the API documentation.
    """

    # 'IWindow' interface -------------------------------------------------#

    title = Str("Pyface")

    # ------------------------------------------------------------------------
    # Protected 'IApplicationWindow' interface.
    # ------------------------------------------------------------------------

    def _create_contents(self, parent):
        panel = wx.Panel(parent, -1, name="ApplicationWindow")
        panel.SetBackgroundColour("blue")
        return panel

    def _create_menu_bar(self, parent):
        if self.menu_bar_manager is not None:
            menu_bar = self.menu_bar_manager.create_menu_bar(parent)
            self.control.SetMenuBar(menu_bar)

    def _create_status_bar(self, parent):
        if self.status_bar_manager is not None:
            status_bar = self.status_bar_manager.create_status_bar(parent)
            self.control.SetStatusBar(status_bar)

    def _create_tool_bar(self, parent):
        tool_bar_managers = self.tool_bar_managers
        if len(tool_bar_managers) > 0:
            for tool_bar_manager in reversed(tool_bar_managers):
                tool_bar = tool_bar_manager.create_tool_bar(parent, aui=True)
                self._add_toolbar_to_aui_manager(tool_bar)
            self._aui_manager.Update()

    def _set_window_icon(self):
        if self.icon is None:
            icon = ImageResource("application.ico")
        else:
            icon = self.icon

        if self.control is not None:
            self.control.SetIcon(icon.create_icon())

    # ------------------------------------------------------------------------
    # 'Window' interface.
    # ------------------------------------------------------------------------

    def _size_default(self):
        """ Trait initialiser. """
        return (800, 600)

    # ------------------------------------------------------------------------
    # Protected 'IWidget' interface.
    # ------------------------------------------------------------------------

    def create(self, parent=None):

        super().create(parent=parent)

        self._aui_manager = PyfaceAuiManager()
        self._aui_manager.SetManagedWindow(self.control)

        # Keep a reference to the AUI Manager in the control because Panes
        # will need to access it in order to lay themselves out
        self.control._aui_manager = self._aui_manager

        contents = self._create_contents(self.control)
        self._aui_manager.AddPane(contents, aui.AuiPaneInfo().CenterPane())

        self._create_trim_widgets(self.control)

        # Updating the AUI manager actually commits all of the pane's added
        # to it (this allows batch updates).
        self._aui_manager.Update()

    def _create_control(self, parent):

        style = (
            wx.DEFAULT_FRAME_STYLE | wx.FRAME_NO_WINDOW_MENU | wx.CLIP_CHILDREN
        )

        control = wx.Frame(
            parent,
            -1,
            self.title,
            style=style,
            size=self.size,
            pos=self.position,
        )

        # Mac/Win needs this, otherwise background color is black
        attr = control.GetDefaultAttributes()
        control.SetBackgroundColour(attr.colBg)

        return control

    def destroy(self):
        if self.control:
            self._aui_manager.UnInit()
        super().destroy()

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _add_toolbar_to_aui_manager(self, tool_bar):
        """ Add a toolbar to the AUI manager. """

        info = self._get_tool_bar_pane_info(tool_bar)
        self._aui_manager.AddPane(tool_bar, info)

    def _get_tool_bar_pane_info(self, tool_bar):
        info = aui.AuiPaneInfo()
        info.Caption(tool_bar.tool_bar_manager.name)
        info.LeftDockable(False)
        info.Name(tool_bar.tool_bar_manager.id)
        info.RightDockable(False)
        info.ToolbarPane()
        info.Top()

        return info

    def _wx_enable_tool_bar(self, tool_bar, enabled):
        """ Enable/Disablea tool bar. """

        # AUI toolbars cannot be enabled/disabled.

    def _wx_show_tool_bar(self, tool_bar, visible):
        """ Hide/Show a tool bar. """

        pane = self._aui_manager.GetPane(tool_bar.tool_bar_manager.id)

        if visible:
            pane.Show()

        else:
            # Without this workaround, toolbars know the sizes of other
            # hidden toolbars and leave gaps in the toolbar dock
            pane.window.Show(False)
            self._aui_manager.DetachPane(pane.window)
            info = self._get_tool_bar_pane_info(pane.window)
            info.Hide()
            self._aui_manager.AddPane(pane.window, info)

        self._aui_manager.Update()

        return

    # Trait change handlers ------------------------------------------------

    def _menu_bar_manager_changed(self):
        if self.control is not None:
            self._create_menu_bar(self.control)

    def _status_bar_manager_changed(self, old, new):
        if self.control is not None:
            if old is not None:
                self.control.SetStatusBar(None)
                old.destroy()
            self._create_status_bar(self.control)

    @observe("tool_bar_managers.items")
    def _update_tool_bar_managers(self, event):
        if self.control is not None:
            self._create_tool_bar(self.control)

    def _icon_changed(self):
        self._set_window_icon()



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traits.api import Array, HasStrictTraits, provides

from pyface.i_image import IImage
from pyface.util.image_helpers import (
    array_to_image, image_to_bitmap, bitmap_to_icon, resize_image
)

#: Trait type for image arrays.
ImageArray = Array(shape=(None, None, (3, 4)), dtype='uint8')


@provides(IImage)
class ArrayImage(HasStrictTraits):
    """ An IImage stored in an RGB(A) numpy array.
    """

    # 'ArrayImage' interface ------------------------------------------------

    #: The bytes of the image.
    data = ImageArray()

    # ------------------------------------------------------------------------
    # 'IImage' interface.
    # ------------------------------------------------------------------------

    def create_image(self, size=None):
        """ Creates a toolkit-specific image for this array.

        Parameters
        ----------
        size : (int, int) or None
            The desired size as a width, height tuple, or None if wanting
            default image size.

        Returns
        -------
        image : toolkit image
            The toolkit image corresponding to the image and the specified
            size.
        """
        image = array_to_image(self.data)
        if size is not None:
            image = resize_image(image, size)
        return image

    def create_bitmap(self, size=None):
        """ Creates a toolkit-specific bitmap image for this array.

        Parameters
        ----------
        size : (int, int) or None
            The desired size as a width, height tuple, or None if wanting
            default image size.

        Returns
        -------
        image : toolkit bitmap
            The toolkit bitmap corresponding to the image and the specified
            size.
        """
        return image_to_bitmap(self.create_image(size))

    def create_icon(self, size=None):
        """ Creates a toolkit-specific icon for this array.

        Parameters
        ----------
        size : (int, int) or None
            The desired size as a width, height tuple, or None if wanting
            default icon size.

        Returns
        -------
        image : toolkit icon
            The toolkit image corresponding to the image and the specified
            size as an icon.
        """
        return bitmap_to_icon(self.create_bitmap(size))

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, data, **traits):
        super().__init__(data=data, **traits)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The interface of a top-level application window. """


from traits.api import HasTraits, Instance, List

from pyface.action.i_menu_bar_manager import IMenuBarManager
from pyface.action.i_status_bar_manager import IStatusBarManager
from pyface.action.i_tool_bar_manager import IToolBarManager
from pyface.i_window import IWindow
from pyface.ui_traits import Image


class IApplicationWindow(IWindow):
    """ The interface for a top-level application window.

    The application window has support for a menu bar, tool bar and a status
    bar (all of which are optional).

    Notes
    -----

    To use, create a sub-class of this class and override the
    :py:meth:`._create_contents` method.
    """

    # 'IApplicationWindow' interface ---------------------------------------

    #: The window icon.  The default is toolkit specific.
    icon = Image()

    #: The menu bar manager for the window.
    menu_bar_manager = Instance(IMenuBarManager)

    #: The status bar manager for the window.
    status_bar_manager = Instance(IStatusBarManager)

    #: The collection of tool bar managers for the window.
    tool_bar_managers = List(Instance(IToolBarManager))

    # ------------------------------------------------------------------------
    # Protected 'IApplicationWindow' interface.
    # ------------------------------------------------------------------------

    def _create_contents(self, parent):
        """ Create and return the window's contents.

        Parameters
        ----------
        parent : toolkit control
            The window's toolkit control to be used as the parent for
            widgets in the contents.

        Returns
        -------
        control : toolkit control
            A control to be used for contents of the window.
        """

    def _create_menu_bar(self, parent):
        """ Creates the menu bar (if required).

        Parameters
        ----------
        parent : toolkit control
            The window's toolkit control.
        """

    def _create_status_bar(self, parent):
        """ Creates the status bar (if required).

        Parameters
        ----------
        parent : toolkit control
            The window's toolkit control.
        """

    def _create_tool_bar(self, parent):
        """ Creates the tool bar (if required).

        Parameters
        ----------
        parent : toolkit control
            The window's toolkit control.
        """

    def _create_trim_widgets(self, parent):
        """ Creates the 'trim' widgets (the widgets around the window).

        Parameters
        ----------
        parent : toolkit control
            The window's toolkit control.
        """

    def _set_window_icon(self):
        """ Sets the window icon (if required). """


class MApplicationWindow(HasTraits):
    """ The mixin class that contains common code for toolkit specific
    implementations of the :py:class:`IApplicationWindow` interface.

    Implements: destroy(), _create_trim_widgets()
    """

    #: The icon to display in the application window title bar.
    icon = Image()

    #: The menu bar manager for the window.
    menu_bar_manager = Instance(IMenuBarManager)

    #: The status bar manager for the window.
    status_bar_manager = Instance(IStatusBarManager)

    #: The collection of tool bar managers for the window.
    tool_bar_managers = List(Instance(IToolBarManager))

    # ------------------------------------------------------------------------
    # 'IWidget' interface.
    # ------------------------------------------------------------------------

    def destroy(self):
        """ Destroy the control if it exists. """

        if self.menu_bar_manager is not None:
            self.menu_bar_manager.destroy()

        if self.status_bar_manager is not None:
            self.status_bar_manager.destroy()

        for tool_bar_manager in self.tool_bar_managers:
            tool_bar_manager.destroy()

        super().destroy()

    # ------------------------------------------------------------------------
    # Protected 'IApplicationWindow' interface.
    # ------------------------------------------------------------------------

    def _create_trim_widgets(self, parent):
        """ Creates the 'trim' widgets (the widgets around the window).

        Parameters
        ----------
        parent : toolkit control
            The window's toolkit control.
        """

        # All of these are optional.
        self._set_window_icon()
        self._create_menu_bar(parent)
        self._create_tool_bar(parent)
        self._create_status_bar(parent)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import wx
from wx.grid import Grid
from wx.grid import GridCellFloatRenderer, GridCellFloatEditor


class AbstractGridView(Grid):
    """ Enthought's default spreadsheet view.

    Uses a virtual data source.

    THIS CLASS IS NOT LIMITED TO ONLY DISPLAYING LOG DATA!
    """

    def __init__(self, parent, ID=-1, **kw):

        Grid.__init__(self, parent, ID, **kw)

        # We have things set up to edit on a single click - so we have to select
        # an initial cursor location that is off of the screen otherwise a cell
        # will be in edit mode as soon as the grid fires up.
        self.moveTo = [1000, 1]
        self.edit = False

        # this seems like a busy idle ...
        self.Bind(wx.EVT_IDLE, self.OnIdle)

        # Enthought specific display controls ...
        self.init_labels()
        self.init_data_types()
        self.init_handlers()

        self.Bind(wx.grid.EVT_GRID_EDITOR_CREATED, self._on_editor_created)

    def init_labels(self):
        self.SetLabelFont(
            wx.Font(
                self.GetFont().GetPointSize(), wx.SWISS, wx.NORMAL, wx.BOLD
            )
        )
        self.SetGridLineColour("blue")
        self.SetColLabelAlignment(wx.ALIGN_CENTRE, wx.ALIGN_CENTRE)
        self.SetRowLabelAlignment(wx.ALIGN_LEFT, wx.ALIGN_CENTRE)

    def init_data_types(self):
        """ If the model says a cell is of a specified type, the grid uses
        the specific renderer and editor set in this method.
        """
        self.RegisterDataType(
            "LogData",
            GridCellFloatRenderer(precision=3),
            GridCellFloatEditor(),
        )

    def init_handlers(self):

        self.Bind(wx.grid.EVT_GRID_CELL_LEFT_CLICK, self.OnCellLeftClick)
        self.Bind(wx.grid.EVT_GRID_CELL_RIGHT_CLICK, self.OnCellRightClick)
        self.Bind(wx.grid.EVT_GRID_CELL_LEFT_DCLICK, self.OnCellLeftDClick)
        self.Bind(wx.grid.EVT_GRID_CELL_RIGHT_DCLICK, self.OnCellRightDClick)

        self.Bind(wx.grid.EVT_GRID_LABEL_LEFT_CLICK, self.OnLabelLeftClick)
        self.Bind(wx.grid.EVT_GRID_LABEL_RIGHT_CLICK, self.OnLabelRightClick)
        self.Bind(wx.grid.EVT_GRID_LABEL_LEFT_DCLICK, self.OnLabelLeftDClick)
        self.Bind(wx.grid.EVT_GRID_LABEL_RIGHT_DCLICK, self.OnLabelRightDClick)

        self.Bind(wx.grid.EVT_GRID_ROW_SIZE, self.OnRowSize)
        self.Bind(wx.grid.EVT_GRID_COL_SIZE, self.OnColSize)

        self.Bind(wx.grid.EVT_GRID_RANGE_SELECT, self.OnRangeSelect)
        self.Bind(wx.grid.EVT_GRID_CELL_CHANGE, self.OnCellChange)
        self.Bind(wx.grid.EVT_GRID_SELECT_CELL, self.OnSelectCell)

        self.Bind(wx.grid.EVT_GRID_EDITOR_SHOWN, self.OnEditorShown)
        self.Bind(wx.grid.EVT_GRID_EDITOR_HIDDEN, self.OnEditorHidden)
        self.Bind(wx.grid.EVT_GRID_EDITOR_CREATED, self.OnEditorCreated)

    def SetColLabelsVisible(self, show=True):
        """ This only works if you 'hide' then 'show' the labels.
        """
        if not show:
            self._default_col_label_size = self.GetColLabelSize()
            self.SetColLabelSize(0)
        else:
            self.SetColLabelSize(self._default_col_label_size)

    def SetRowLabelsVisible(self, show=True):
        """ This only works if you 'hide' then 'show' the labels.
        """
        if not show:
            self._default_row_label_size = self.GetRowLabelSize()
            self.SetRowLabelSize(0)
        else:
            self.SetRowLabelSize(self._default_row_label_size)

    def SetTable(self, table, *args):
        """ Some versions of wxPython do not return the correct
        table - hence we store our own copy here - weak ref?
        todo - does this apply to Enthought?
        """
        self._table = table
        return Grid.SetTable(self, table, *args)

    def GetTable(self):
        # Terminate editing of the current cell to force an update of the table
        self.DisableCellEditControl()
        return self._table

    def Reset(self):
        """ Resets the view based on the data in the table.

        Call this when rows are added or destroyed.
        """
        self._table.ResetView(self)

    def OnCellLeftClick(self, evt):
        evt.Skip()

    def OnCellRightClick(self, evt):
        # print self.GetDefaultRendererForCell(evt.GetRow(), evt.GetCol())
        evt.Skip()

    def OnCellLeftDClick(self, evt):
        if self.CanEnableCellControl():
            self.EnableCellEditControl()
        evt.Skip()

    def OnCellRightDClick(self, evt):
        evt.Skip()

    def OnLabelLeftClick(self, evt):
        evt.Skip()

    def OnLabelRightClick(self, evt):
        evt.Skip()

    def OnLabelLeftDClick(self, evt):
        evt.Skip()

    def OnLabelRightDClick(self, evt):
        evt.Skip()

    def OnRowSize(self, evt):
        evt.Skip()

    def OnColSize(self, evt):
        evt.Skip()

    def OnRangeSelect(self, evt):
        # if evt.Selecting():
        #    print "OnRangeSelect: top-left %s, bottom-right %s\n" % (evt.GetTopLeftCoords(), evt.GetBottomRightCoords())
        evt.Skip()

    def OnCellChange(self, evt):
        evt.Skip()

    def OnIdle(self, evt):
        """ Immediately jumps into editing mode, bypassing the usual select mode
        of a spreadsheet. See also self.OnSelectCell().
        """

        if self.edit:
            if self.CanEnableCellControl():
                self.EnableCellEditControl()
            self.edit = False

        if self.moveTo is not None:
            self.SetGridCursor(self.moveTo[0], self.moveTo[1])
            self.moveTo = None

        evt.Skip()

    def OnSelectCell(self, evt):

        """ Immediately jumps into editing mode, bypassing the usual select mode
        of a spreadsheet. See also self.OnIdle().
        """
        self.edit = True
        evt.Skip()

    def OnEditorShown(self, evt):
        evt.Skip()

    def OnEditorHidden(self, evt):
        evt.Skip()

    def OnEditorCreated(self, evt):
        evt.Skip()



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import wx

from traits.api import provides

from pyface.i_layout_item import DEFAULT_SIZE
from pyface.i_layout_widget import ILayoutWidget, MLayoutWidget
from pyface.ui.wx.widget import Widget


#: The special "default" size for Wx Size objects.
WX_DEFAULT_SIZE = -1


@provides(ILayoutWidget)
class LayoutWidget(MLayoutWidget, Widget):
    """ A widget which can participate as part of a layout.

    This is an abstract class, as Widget._create_control needs to be
    implemented at a minimum.
    """

    def _set_control_minimum_size(self, size):
        """ Set the minimum size of the control. """
        wx_size = _size_to_wx_size(size)
        self.control.SetMinSize(wx_size)

    def _get_control_minimum_size(self):
        """ Get the minimum size of the control. """
        wx_size = self.control.GetMinSize()
        return _wx_size_to_size(wx_size)

    def _set_control_maximum_size(self, size):
        """ Set the maximum size of the control. """
        wx_size = _size_to_wx_size(size)
        self.control.SetMaxSize(wx_size)

    def _get_control_maximum_size(self):
        """ Get the maximum size of the control. """
        wx_size = self.control.GetMaxSize()
        return _wx_size_to_size(wx_size)

    def _set_control_stretch(self, size_policy):
        """ Set the stretch factor of the control.

        In Wx the stretch factor is set at layout time and can't be changed
        without removing and adding the widget back into a layout.
        """
        pass

    def _get_control_stretch(self):
        """ Get the stretch factor of the control.

        In Wx the stretch factor is set at layout time and can't be obtained
        from the widget itself.  As a result, this method simply returns the
        current trait value.  This method is only used for testing.
        """
        return self.stretch

    def _set_control_size_policy(self, size_policy):
        """ Set the size policy of the control

        In Wx the size policy is set at layout time and can't be changed
        without removing and adding the widget back into a layout.
        """
        pass

    def _get_control_size_policy(self):
        """ Get the size policy of the control

        In Wx the size policy is set at layout time and can't be obtained from
        the widget itself.  As a result, this method simply returns the
        current trait value.  This method is only used for testing.
        """
        return self.size_policy


def _size_to_wx_size(size):
    """ Convert a size tuple to a wx.Size instance.

    Parameters
    ----------
    size : tuple of (width, height)
        The width and height as a tuple of ints.

    Returns
    -------
    wx_size : wx.Size
        A corresponding wx Size instance.
    """
    return wx.Size(*(
        x if x != DEFAULT_SIZE else WX_DEFAULT_SIZE
        for x in size
    ))


def _wx_size_to_size(wx_size):
    """ Convert a wx.Size instance to a size tuple.

    Parameters
    ----------
    wx_size : wx.Size
        A wx Size instance.

    Returns
    -------
    size : tuple of (width, height)
        The corresponding width and height as a tuple of ints.
    """
    return (wx_size.GetWidth(), wx_size.GetHeight())



# Copyright (c) 2005-2023, Enthought Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!

import unittest


class TestImageHelpers(unittest.TestCase):

    def test_imports(self):
        # actual functions are tested in toolkits
        from ..image_helpers import (  # noqa: F401
            AspectRatio, ScaleMode, array_to_image, bitmap_to_icon,
            bitmap_to_image, image_to_array, image_to_bitmap,
            resize_bitmap, resize_image,
        )



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest
from unittest.mock import MagicMock

from pyface.toolkit import toolkit

not_wx = toolkit.toolkit != "wx"


class TestDockControl(unittest.TestCase):

    @unittest.skipIf(not_wx, "This test is specific to the wx backend")
    def test_feature_changed(self):
        from pyface.dock.dock_sizer import DockControl
        dock_control = DockControl()
        DockControl.set_feature_mode = MagicMock()
        dock_control.feature_changed = True

        dock_control.set_feature_mode.assert_called_once_with()



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traits.api import provides

from pyface.i_pil_image import IPILImage, MPILImage
from pyface.ui.qt.util.image_helpers import resize_image


@provides(IPILImage)
class PILImage(MPILImage):
    """ The toolkit specific implementation of a PILImage.
    """

    # ------------------------------------------------------------------------
    # 'IImage' interface.
    # ------------------------------------------------------------------------

    def create_image(self, size=None):
        """ Creates a Qt image for this image.

        Parameters
        ----------
        size : (int, int) or None
            The desired size as a (width, height) tuple, or None if wanting
            default image size.

        Returns
        -------
        image : QImage
            The toolkit image corresponding to the image and the specified
            size.
        """
        from PIL.ImageQt import ImageQt
        image = ImageQt(self.image)
        if size is not None:
            return resize_image(image, size)
        else:
            return image



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>

""" Utilities for handling alignment """

import wx


ALIGNMENT_TO_WX_ALIGNMENT = {
    'default': wx.TE_LEFT,
    'left': wx.TE_LEFT,
    'center': wx.TE_CENTRE,
    'right': wx.TE_RIGHT,
}
WX_ALIGNMENT_TO_ALIGNMENT = {
    0: 'default',
    wx.TE_LEFT: 'left',
    wx.TE_CENTRE: 'center',
    wx.TE_RIGHT: 'right',
}
ALIGNMENT_MASK = wx.TE_LEFT | wx.TE_CENTRE | wx.TE_RIGHT


def get_alignment_style(style_flags):
    alignment_flag = style_flags & ALIGNMENT_MASK
    return WX_ALIGNMENT_TO_ALIGNMENT[alignment_flag]


def set_alignment_style(alignment, style_flags):
    other_flags = style_flags & ~ALIGNMENT_MASK
    return other_flags | ALIGNMENT_TO_WX_ALIGNMENT[alignment]



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Helper functions for working with images

This module provides helper functions for converting between numpy arrays
and toolkit images, as well as between the various image types in a standardized
way.

Helper functions
----------------

- :data:`~.array_to_image`
- :data:`~.bitmap_to_icon`
- :data:`~.bitmap_to_image`
- :data:`~.image_to_array`
- :data:`~.image_to_bitmap`
- :data:`~.resize_image`
- :data:`~.resize_bitmap`

Options for resizing images
---------------------------
- :data:`~.ScaleMode`
- :data:`~.AspectRatio`

"""

from pyface.toolkit import toolkit_object

# Enum types for function arguments
ScaleMode = toolkit_object("util.image_helpers:ScaleMode")
AspectRatio = toolkit_object("util.image_helpers:AspectRatio")

# Helper functions
array_to_image = toolkit_object("util.image_helpers:array_to_image")
bitmap_to_icon = toolkit_object("util.image_helpers:bitmap_to_icon")
bitmap_to_image = toolkit_object("util.image_helpers:bitmap_to_image")
image_to_array = toolkit_object("util.image_helpers:image_to_array")
image_to_bitmap = toolkit_object("util.image_helpers:image_to_bitmap")
resize_image = toolkit_object("util.image_helpers:resize_image")
resize_bitmap = toolkit_object("util.image_helpers:resize_bitmap")



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from enum import Enum

from traits.api import provides

from pyface.qt import QtGui
from pyface.i_layout_item import DEFAULT_SIZE
from pyface.i_layout_widget import ILayoutWidget, MLayoutWidget
from pyface.ui.qt.widget import Widget


#: Maximum widget size (some versions of PyQt don't export it)
QWIDGETSIZE_MAX = getattr(QtGui, "QWIDGETSIZE_MAX", 1 << 24 - 1)


class SizePolicies(Enum):
    """ Qt values for size policies

    Note that Qt has additional values that are not mapped to Pyface size
    policies.
    """
    fixed = QtGui.QSizePolicy.Policy.Fixed
    preferred = QtGui.QSizePolicy.Policy.Preferred
    expand = QtGui.QSizePolicy.Policy.Expanding


@provides(ILayoutWidget)
class LayoutWidget(MLayoutWidget, Widget):
    """ A widget which can participate as part of a layout. """

    def _set_control_minimum_size(self, size):
        size = tuple(
            x if x != DEFAULT_SIZE else 0
            for x in size
        )
        self.control.setMinimumSize(*size)

    def _get_control_minimum_size(self):
        size = self.control.minimumSize()
        return (size.width(), size.height())

    def _set_control_maximum_size(self, size):
        size = tuple(
            x if x != DEFAULT_SIZE else QWIDGETSIZE_MAX
            for x in size
        )
        self.control.setMaximumSize(*size)

    def _get_control_maximum_size(self):
        size = self.control.maximumSize()
        return (size.width(), size.height())

    def _set_control_stretch(self, stretch):
        """ Set the stretch factor of the control.
        """
        new_size_policy = _clone_size_policy(self.control.sizePolicy())
        new_size_policy.setHorizontalStretch(stretch[0])
        new_size_policy.setVerticalStretch(stretch[1])
        self.control.setSizePolicy(new_size_policy)

    def _get_control_stretch(self):
        """ Get the stretch factor of the control.

        This method is only used for testing.
        """
        size_policy = self.control.sizePolicy()
        return (
            size_policy.horizontalStretch(),
            size_policy.verticalStretch(),
        )

    def _set_control_size_policy(self, size_policy):
        new_size_policy = _clone_size_policy(self.control.sizePolicy())
        if size_policy[0] != "default":
            new_size_policy.setHorizontalPolicy(
                SizePolicies[size_policy[0]].value
            )
        if size_policy[1] != "default":
            new_size_policy.setVerticalPolicy(
                SizePolicies[size_policy[1]].value
            )
        self.control.setSizePolicy(new_size_policy)

    def _get_control_size_policy(self):
        size_policy = self.control.sizePolicy()
        if self.size_policy[0] != "default":
            horizontal_policy = SizePolicies(
                size_policy.horizontalPolicy()).name
        else:
            horizontal_policy = "default"
        if self.size_policy[1] != "default":
            vertical_policy = SizePolicies(
                size_policy.verticalPolicy()).name
        else:
            vertical_policy = "default"
        return (horizontal_policy, vertical_policy)

    def destroy(self):
        if self.control is not None:
            self.control.hide()
            super().destroy()


def _clone_size_policy(size_policy):
    """ Clone the state of an existing QSizePolicy object

    This is required because there is no standard Qt copy or clone
    method.
    """
    new_size_policy = QtGui.QSizePolicy()
    new_size_policy.setHorizontalPolicy(
        size_policy.horizontalPolicy()
    )
    new_size_policy.setVerticalPolicy(
        size_policy.verticalPolicy()
    )
    new_size_policy.setHorizontalStretch(
        size_policy.horizontalStretch()
    )
    new_size_policy.setVerticalStretch(
        size_policy.verticalStretch()
    )
    new_size_policy.setHeightForWidth(
        size_policy.hasHeightForWidth()
    )
    new_size_policy.setWidthForHeight(
        size_policy.hasWidthForHeight()
    )
    return new_size_policy



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Base action manager builder class

This module provides a base class that takes a schema for an action manager
and builds the concrete action manager and its groups and items, folding in
schema additions.
"""

import logging

from collections import defaultdict

from traits.api import HasTraits, Instance, List

from .schema import Schema, ToolBarSchema
from .schema_addition import SchemaAddition
from ._topological_sort import before_after_sort


# Logging.
logger = logging.getLogger(__name__)


class ActionManagerBuilder(HasTraits):
    """ Builds action managers from schemas, merging schema additions.
    """

    #: An ActionController to use with the action managers created by the
    #: builder.  May be None.
    controller = Instance("pyface.action.action_controller.ActionController")

    #: Schema additions to apply to all managers built by this builder.
    #: Any additions which do not match part of the supplied schema will be
    #: ignored.
    additions = List(Instance(SchemaAddition))

    # ------------------------------------------------------------------------
    # 'ActionManagerBuilder' interface.
    # ------------------------------------------------------------------------

    def create_action_manager(self, schema):
        """ Create a manager for the given schema using the additions.

        Any additions whose paths do not match the supplied

        Parameters
        ----------
        schema : Schema
            An Schema for an ActionManager subclass (ie. one of MenuBarSchema,
            MenuSchema, or ToolBarSchema).

        Returns
        -------
        manager : ActionManager
            The concrete ActionManager instance corresponding to the Schema
            with addtions.  This does not yet have concrete toolkit widgets
            associated with it: usually those will be created separately.
        """
        additions_map = defaultdict(list)
        for addition in self.additions:
            if addition.path:
                additions_map[addition.path].append(addition)

        manager = self._create_action_manager_recurse(schema, additions_map)
        manager.controller = self.controller
        return manager

    def get_additional_toolbar_schemas(self):
        """ Get any top-level toolbars from additions.

        Unlike menus, there is no base toolbar manager, so additions which
        contribute new toolbars appear with no path.  It is up to the class
        using the builder how it wants to handle these additional toolbars.

        Returns
        -------
        schemas : list of ToolBarSchema
            The additional toolbars specified in self.additions.
        """
        schemas = []
        for addition in self.additions:
            if not addition.path:
                schema = addition.factory()
                if isinstance(schema, ToolBarSchema):
                    schemas.append(schema)
                else:
                    logger.error(
                        "Invalid top-level schema addition: %r. Only "
                        "ToolBar schemas can be path-less.",
                        schema,
                    )
        return schemas

    def prepare_item(self, item, path):
        """ Called immediately after a concrete Pyface item has been created
        (or, in the case of items that are not produced from schemas,
        immediately before they are processed).

        This hook can be used to perform last-minute transformations or
        configuration. Returns a concrete Pyface item.

        Parameters
        ----------
        item : pyface.action item
            A concrete Pyface item.
        path : str
            The path to the item in the Schema.

        Returns
        -------
        item : pyface.action item
            A modified or transformed Pyface item.
        """
        return item

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _get_ordered_schemas(self, schemas):
        begin = []
        middle = []
        end = []

        for schema in schemas:
            absolute_position = getattr(schema, "absolute_position", None)
            if absolute_position is None:
                middle.append(schema)
            elif absolute_position == "last":
                end.append(schema)
            else:
                begin.append(schema)

        schemas = (
            before_after_sort(begin)
            + before_after_sort(middle)
            + before_after_sort(end)
        )
        return schemas

    def _group_items_by_id(self, items):
        """ Group a list of action items by their ID.

        Action items are Schemas and Groups, MenuManagers, etc.

        Return a dictionary {item_id: list_of_items}, and a list containing
        all the ids ordered by their appearance in the `all_items` list. The
        ordered IDs are used as a replacement for an ordered dictionary, to
        keep compatibility with Python <2.7 .

        """

        ordered_items_ids = []
        id_to_items = defaultdict(list)

        for item in items:
            if item.id not in id_to_items:
                ordered_items_ids.append(item.id)
            id_to_items[item.id].append(item)

        return id_to_items, ordered_items_ids

    def _group_items_by_class(self, items):
        """ Group a list of action items by their class.

        Action items are Schemas and Groups, MenuManagers, etc.

        Return a dictionary {item_class: list_of_items}, and a list containing
        all the classes ordered by their appearance in the `all_items` list.
        The ordered classes are used as a replacement for an ordered
        dictionary, to keep compatibility with Python <2.7 .

        """

        ordered_items_class = []
        class_to_items = defaultdict(list)

        for item in items:
            if item.__class__ not in class_to_items:
                ordered_items_class.append(item.__class__)
            class_to_items[item.__class__].append(item)

        return class_to_items, ordered_items_class

    def _unpack_schema_additions(self, items):
        """ Unpack additions, since they may themselves be schemas. """

        unpacked_items = []

        for item in items:
            if isinstance(item, SchemaAddition):
                unpacked_items.append(item.factory())
            else:
                unpacked_items.append(item)

        return unpacked_items

    def _merge_items_with_same_path(self, id_to_items, ordered_items_ids):
        """ Merge items with the same path if possible.

        Items must be subclasses of `Schema` and they must be instances of
        the same class to be merged.

        """

        merged_items = []
        for item_id in ordered_items_ids:
            items_with_same_id = id_to_items[item_id]

            # Group items by class.
            class_to_items, ordered_items_class = self._group_items_by_class(
                items_with_same_id
            )

            for items_class in ordered_items_class:
                items_with_same_class = class_to_items[items_class]

                if len(items_with_same_class) == 1:
                    merged_items.extend(items_with_same_class)

                else:
                    # Only schemas can be merged.
                    if issubclass(items_class, Schema):
                        # Merge into a single schema.
                        items_content = sum(
                            (item.items for item in items_with_same_class), []
                        )

                        merged_item = items_with_same_class[0].clone_traits()
                        merged_item.items = items_content
                        merged_items.append(merged_item)

                    else:
                        merged_items.extend(items_with_same_class)

        return merged_items

    def _preprocess_schemas(self, schema, additions, path):
        """ Sort and merge a schema and a set of schema additions. """

        # Determine the order of the items at this path.
        if additions[path]:
            all_items = self._get_ordered_schemas(
                schema.items + additions[path]
            )
        else:
            all_items = schema.items

        unpacked_items = self._unpack_schema_additions(all_items)

        id_to_items, ordered_items_ids = self._group_items_by_id(
            unpacked_items
        )

        merged_items = self._merge_items_with_same_path(
            id_to_items, ordered_items_ids
        )

        return merged_items

    def _create_action_manager_recurse(self, schema, additions, path=""):
        """ Recursively create a manager for the given schema and additions map.

        Items with the same path are merged together in a single entry if
        possible (i.e., if they have the same class).

        When a list of items is merged, their children are added to a clone
        of the first item in the list. As a consequence, traits like menu
        names etc. are inherited from the first item.

        """
        from pyface.action.action_manager import ActionManager

        # Compute the new action path.
        if path:
            path = path + "/" + schema.id
        else:
            path = schema.id

        preprocessed_items = self._preprocess_schemas(schema, additions, path)

        # Create the actual children by calling factory items.
        children = []
        for item in preprocessed_items:
            if isinstance(item, Schema):
                item = self._create_action_manager_recurse(
                    item, additions, path
                )
            else:
                item = self.prepare_item(item, path + "/" + item.id)

            if isinstance(item, ActionManager):
                # Give even non-root action managers a reference to the
                # controller so that custom Groups, MenuManagers, etc. can get
                # access to the state it holds.
                item.controller = self.controller

            children.append(item)

        # Finally, create the pyface.action instance for this schema.
        return self.prepare_item(schema.create(children), path)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A header for an entry in a collection of expandables.

The header provides a visual indicator of the current state, a text label,
and a 'remove' button.
"""

import warnings

import wx

from traits.api import Event, Str, Bool

from pyface.wx.util.font_helper import new_font_like
from pyface.ui_traits import Image
from .image_resource import ImageResource
from .widget import Widget


class ExpandableHeader(Widget):
    """ A header for an entry in a collection of expandables.

    The header provides a visual indicator of the current state, a text label,
    and a 'remove' button.
    """

    #: The title of the panel.
    title = Str("Panel")

    #: The carat image to show when the panel is collapsed.
    collapsed_carat_image = Image(ImageResource("carat_closed"))

    #: The carat image to show when the panel is expanded.
    expanded_carat_image = Image(ImageResource("carat_open"))

    #: Represents the current state of the panel. True means expanded.
    state = Bool(False)

    # Events ----

    #: The panel has been expanded or collapsed
    panel_expanded = Event()

    #: The panel has been closed
    panel_closed = Event()

    _CARAT_X = 4
    _CARAT_Y = 4
    _TEXT_Y = 0
    _TEXT_X_OFFSET = 10

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, parent=None, container=None, **traits):
        """ Creates the panel. """

        if container is not None:
            warnings.warn(
                "the container parameter is deprecated and will be "
                "removed in a future Pyface release",
                DeprecationWarning,
                stacklevel=2,
            )
            self.observe(
                lambda event: container.remove_panel(event.new),
                "panel_closed",
            )
        create = traits.pop("create", None)

        # Base class constructor.
        super().__init__(parent=parent, **traits)

        # Create the toolkit-specific control that represents the widget.
        if create:
            self.create()
            warnings.warn(
                "automatic widget creation is deprecated and will be removed "
                "in a future Pyface version, code should not pass the create "
                "parameter and should instead call create() explicitly",
                DeprecationWarning,
                stacklevel=2,
            )
        elif create is not None:
            warnings.warn(
                "setting create=False is no longer required",
                DeprecationWarning,
                stacklevel=2,
            )

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _create_control(self, parent):
        """ Create the toolkit-specific control that represents the widget. """

        collapsed_carat = self.collapsed_carat_image.create_image()
        self._collapsed_bmp = collapsed_carat.ConvertToBitmap()
        self._carat_w = self._collapsed_bmp.GetWidth()

        expanded_carat = self.expanded_carat_image.create_image()
        self._expanded_bmp = expanded_carat.ConvertToBitmap()

        # create our panel and initialize it appropriately
        sizer = wx.BoxSizer(wx.VERTICAL)
        panel = wx.Panel(parent, -1, style=wx.CLIP_CHILDREN | wx.BORDER_SIMPLE)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)

        # create the remove button
        remove = wx.BitmapButton.NewCloseButton(panel, -1)
        sizer.Add(remove, 0, wx.ALIGN_RIGHT, 5)

        # Create a suitable font.
        self._font = new_font_like(
            wx.NORMAL_FONT, point_size=wx.NORMAL_FONT.GetPointSize() - 1
        )

        height = self._get_preferred_height(parent, self.title, self._font)
        panel.SetMinSize((-1, height+2))

        panel.Bind(wx.EVT_PAINT, self._on_paint)
        panel.Bind(wx.EVT_LEFT_DOWN, self._on_down)
        panel.Bind(wx.EVT_BUTTON, self._on_remove)

        return panel

    def _get_preferred_height(self, parent, text, font):
        """ Calculates the preferred height of the widget. """

        dc = wx.MemoryDC()

        dc.SetFont(font)
        metrics = dc.GetFontMetrics()
        text_h = metrics.height + 2 * self._TEXT_Y

        # add in height of buttons
        carat_h = self._collapsed_bmp.GetHeight() + 2 * self._CARAT_Y

        return max(text_h, carat_h)

    def _draw_carat_button(self, dc):
        """ Draws the button at the correct coordinates. """

        if self.state:
            bmp = self._expanded_bmp
        else:
            bmp = self._collapsed_bmp

        dc.DrawBitmap(bmp, self._CARAT_X, self._CARAT_Y, True)

    def _draw_title(self, dc):
        """ Draws the text label for the header. """
        dc.SetFont(self._font)

        # Render the text.
        dc.DrawText(
            self.title, self._carat_w + self._TEXT_X_OFFSET, self._TEXT_Y
        )

    def _draw(self, dc):
        """ Draws the control. """

        # Draw the title text
        self._draw_title(dc)

        # Draw the carat button
        self._draw_carat_button(dc)

    # ------------------------------------------------------------------------
    # wx event handlers.
    # ------------------------------------------------------------------------

    def _on_paint(self, event):
        """ Called when the background of the panel is erased. """

        # print('ImageButton._on_erase_background')
        dc = wx.PaintDC(self.control)
        self._draw(dc)

    def _on_down(self, event):
        """ Called when button is pressed. """

        self.state = not self.state
        self.control.Refresh()

        # fire an event so any listeners can pick up the change
        self.panel_expanded = self
        event.Skip()

    def _on_remove(self, event):
        """ Called when remove button is pressed. """
        self.panel_closed = self



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Interace and mixins for layered panels.

A layered panel contains one or more named layers, with only one layer
visible at any one time (think of a 'tab' control minus the tabs!).
"""

import warnings

from traits.api import Any, Dict, HasTraits, Interface, Str


class ILayeredPanel(Interface):
    """ A Layered panel.

    A layered panel contains one or more named layers, with only one layer
    visible at any one time (think of a 'tab' control minus the tabs!).  Each
    layer is a toolkit-specific control.

    """

    # "ILayeredPanel' interface --------------------------------------------

    # The toolkit-specific control of the currently displayed layer.
    current_layer = Any()

    # The name of the currently displayed layer.
    current_layer_name = Str()

    # ------------------------------------------------------------------------
    # 'ILayeredPanel' interface.
    # ------------------------------------------------------------------------

    def add_layer(self, name, layer):
        """ Adds a layer with the specified name.

        All layers are hidden when they are added.  Use 'show_layer' to make a
        layer visible.

        """

    def show_layer(self, name):
        """ Shows the layer with the specified name. """

    def has_layer(self, name):
        """ Does the panel contain a layer with the specified name? """


class MLayeredPanel(HasTraits):
    """ A Layered panel mixin.

    A layered panel contains one or more named layers, with only one layer
    visible at any one time (think of a 'tab' control minus the tabs!).  Each
    layer is a toolkit-specific control.
    """

    # "ILayeredPanel' interface --------------------------------------------

    # The toolkit-specific control of the currently displayed layer.
    current_layer = Any()

    # The name of the currently displayed layer.
    current_layer_name = Str()

    # Private traits -------------------------------------------------------

    # The a map of layer names to toolkit controls in the panel.
    _layers = Dict(Str)

    # ------------------------------------------------------------------------
    # 'ILayeredPanel' interface.
    # ------------------------------------------------------------------------

    def has_layer(self, name):
        """ Does the panel contain a layer with the specified name? """
        return name in self._layers

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, parent=None, **traits):
        """ Creates a new LayeredPanel. """

        create = traits.pop("create", None)

        # Base class constructor.
        super().__init__(parent=parent, **traits)

        if create:
            # Create the toolkit-specific control that represents the widget.
            self.create()
            warnings.warn(
                "automatic widget creation is deprecated and will be removed "
                "in a future Pyface version, code should not pass the create "
                "parameter and should instead call create() explicitly",
                DeprecationWarning,
                stacklevel=2,
            )
        elif create is not None:
            warnings.warn(
                "setting create=False is no longer required",
                DeprecationWarning,
                stacklevel=2,
            )



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""

API for the ``pyface`` package.

- :class:`~.Application`
- :class:`~.ApplicationWindow`
- :attr:`~.clipboard`
- :class:`~.Clipboard`
- :func:`~.find_toolkit`
- :class:`~.GUI`
- :class:`~.GUIApplication`
- :class:`~.ImageResource`
- :class:`~.KeyPressedEvent`
- :class:`~.SplashScreen`
- :class:`~.SplitApplicationWindow`
- :class:`~.SplitPanel`
- :class:`~.SystemMetrics`
- :class:`~.Toolkit`
- :class:`~.Window`
- :class:`~.Widget`

Dialogs
-------

- :class:`~.AboutDialog`
- :class:`~.confirm`
- :class:`~.ConfirmationDialog`
- :class:`~.Dialog`
- :class:`~.DirectoryDialog`
- :class:`~.FileDialog`
- :class:`~.error`
- :class:`~.information`
- :class:`~.warning`
- :class:`~.MessageDialog`
- :class:`~.ProgressDialog`
- :class:`~.choose_one`
- :class:`~.SingleChoiceDialog`
- :class:`~.SplitDialog`

Interfaces
----------

- :class:`~.IAboutDialog`
- :class:`~.IApplicationWindow`
- :class:`~.IClipboard`
- :class:`~.IConfirmationDialog`
- :class:`~.IDialog`
- :class:`~.IDirectoryDialog`
- :class:`~.IDropHandler`
- :class:`~.IFileDialog`
- :class:`~.IGUI`
- :class:`~.IHeadingText`
- :class:`~.IImage`
- :class:`~.IImageResource`
- :class:`~.ILayeredPanel`
- :class:`~.ILayoutItem`
- :class:`~.ILayoutWidget`
- :class:`~.IMessageDialog`
- :class:`~.IPILImage`
- :class:`~.IProgressDialog`
- :class:`~.IPythonEditor`
- :class:`~.IPythonShell`
- :class:`~.ISingleChoiceDialog`
- :class:`~.ISplashScreen`
- :class:`~.ISplitWidget`
- :class:`~.ISystemMetrics`
- :class:`~.IWidget`
- :class:`~.IWindow`

Constants
---------

- :class:`~.OK`
- :class:`~.CANCEL`
- :class:`~.YES`
- :class:`~.NO`

UI Traits
---------

- :attr:`~.Alignment`
- :class:`~.HasBorder`
- :class:`~.HasMargin`
- :class:`~.Image`
- :attr:`~.Orientation`
- :attr:`~.Position`
- :class:`~.PyfaceColor`
- :class:`~.PyfaceFont`

Miscellaneous
-------------

- :class:`~.ArrayImage`
- :class:`~.BaseDropHandler`
- :class:`~.Border`
- :class:`~.beep`
- :class:`~.FileDropHandler`
- :class:`~.Filter`
- :class:`~.HeadingText`
- :class:`~.ImageCache`
- :class:`~.LayeredPanel`
- :class:`~.Margin`
- :class:`~.PILImage`
- :class:`~.PythonEditor`
- :class:`~.PythonShell`
- :class:`~.Sorter`

Note that the :class:`~.ArrayImage` is only available if the ``numpy``
package is available in the Python environment.

Note that the :class:`~.PILImage` is only available if the ``pillow``
package is available in the Python environment.

Note that the :class:`~.PythonEditor` and :class:`~.PythonShell` classes are
only available if the ``pygments`` package is available in the Python
environment.

Wx-specific classes
-------------------

- :class:`~.ExpandablePanel`
- :class:`~.ImageWidget`
- :class:`~.MDIApplicationWindow`
- :class:`~.MDIWindowMenu`
- :class:`~.MultiToolbarWindow`

"""

# Imports which don't select the toolkit as a side-effect.

from .application import Application
from .base_toolkit import Toolkit, find_toolkit
from .color import Color
from .constant import OK, CANCEL, YES, NO
from .filter import Filter
from .font import Font
from .gui_application import GUIApplication
from .i_about_dialog import IAboutDialog
from .i_application_window import IApplicationWindow
from .i_clipboard import IClipboard
from .i_confirmation_dialog import IConfirmationDialog
from .i_dialog import IDialog
from .i_directory_dialog import IDirectoryDialog
from .i_drop_handler import IDropHandler
from .i_file_dialog import IFileDialog
from .i_gui import IGUI
from .i_heading_text import IHeadingText
from .i_image import IImage
from .i_image_resource import IImageResource
from .i_layered_panel import ILayeredPanel
from .i_layout_item import ILayoutItem
from .i_layout_widget import ILayoutWidget
from .i_message_dialog import IMessageDialog
from .i_pil_image import IPILImage
from .i_progress_dialog import IProgressDialog
from .i_python_editor import IPythonEditor
from .i_python_shell import IPythonShell
from .i_single_choice_dialog import ISingleChoiceDialog
from .i_splash_screen import ISplashScreen
from .i_split_widget import ISplitWidget
from .i_system_metrics import ISystemMetrics
from .i_widget import IWidget
from .i_window import IWindow
from .sorter import Sorter
from .ui_traits import (
    Alignment, Border, HasBorder, HasMargin, Image, Margin, Orientation,
    Position, PyfaceColor, PyfaceFont
)

# ----------------------------------------------------------------------------
# Deferred imports
# ----------------------------------------------------------------------------

# These imports have the side-effect of performing toolkit selection

_toolkit_imports = {
    'AboutDialog': 'about_dialog',
    'ApplicationWindow': "application_window",
    'BaseDropHandler': "drop_handler",
    'beep': "beep",
    'Clipboard': "clipboard",
    'ConfirmationDialog': "confirmation_dialog",
    'Dialog': "dialog",
    'DirectoryDialog': "directory_dialog",
    'FileDropHandler': "drop_handler",
    'FileDialog': "file_dialog",
    'GUI': 'gui',
    'HeadingText': "heading_text",
    'ImageCache': "image_cache",
    'ImageResource': "image_resource",
    'KeyPressedEvent': "key_pressed_event",
    'LayeredPanel': "layered_panel",
    'MessageDialog': "message_dialog",
    'ProgressDialog': "progress_dialog",
    'SingleChoiceDialog': "single_choice_dialog",
    'SplashScreen': "splash_screen",
    'SystemMetrics': "system_metrics",
    'Window': "window",
    'Widget': "widget",

    # Wx-only (or legacy) imports
    'ExpandablePanel': "expandable_panel",
    'ImageWidget': "image_widget",
    'MDIApplicationWindow': "mdi_application_window",
    'MDIWindowMenu': "mdi_window_menu",
    'MultiToolbarWindow': "multi_toolbar_window",
}

# These are pyface.* imports that have selection as a side-effect
# TODO: refactor to delay imports where possible
_relative_imports = {
    'choose_one': "single_choice_dialog",
    'clipboard': "clipboard",
    'confirm': "confirmation_dialog",
    'error': "message_dialog",
    'information': "message_dialog",
    'SplitApplicationWindow': "split_application_window",
    'SplitDialog': "split_dialog",
    'SplitPanel': "split_panel",
    'warning': "message_dialog",
}
_optional_imports = {
    'ArrayImage': ("numpy", "array_image"),
    'PILImage': ("pillow", "pil_image"),
    'PythonEditor': ("pygments", "python_editor"),
    'PythonShell': ("pygments", "python_shell"),
}


def __getattr__(name):
    """Lazily load attributes with side-effects

    In particular, lazily load toolkit backend names.  For efficiency, lazily
    loaded objects are injected into the module namespace
    """
    # sentinel object for no result
    not_found = object()
    result = not_found

    if name in _relative_imports:
        from importlib import import_module
        source = _relative_imports[name]
        module = import_module(f"pyface.{source}")
        result = getattr(module, name)

    elif name in _toolkit_imports:
        from pyface.toolkit import toolkit_object
        source = _toolkit_imports[name]
        result = toolkit_object(f"{source}:{name}")

    elif name in _optional_imports:
        from importlib import import_module
        import logging
        from pyface.util._optional_dependencies import optional_import
        dependency, source = _optional_imports[name]
        with optional_import(
            dependency,
            msg=f"{name} is not available due to missing {dependency}.",
            logger=logging.getLogger(__name__),
        ):
            module = import_module(f"pyface.{source}")
            result = getattr(module, name)

    if result is not_found:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

    globals()[name] = result
    return result


# ----------------------------------------------------------------------------
# Introspection support
# ----------------------------------------------------------------------------

# the list of available names we report for introspection purposes
_extra_names = (
    set(_toolkit_imports) | set(_relative_imports) | set(_optional_imports)
)


def __dir__():
    return sorted(set(globals()) | _extra_names)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest


from pyface.action.schema.api import SMenu, SMenuBar, SGroup
from pyface.tasks.action.api import DockPaneToggleGroup
from pyface.tasks.api import DockPane, Task, TaskPane, TaskWindow
from pyface.gui import GUI
from traits.api import List
from traits.etsconfig.api import ETSConfig


USING_WX = ETSConfig.toolkit not in {"", "qt", "qt4"}


class BogusTask(Task):

    id = "tests.bogus_task"
    name = "Bogus Task"

    dock_panes = List()

    def create_central_pane(self):
        return TaskPane(id="tests.bogus_task.central_pane")

    def create_dock_panes(self):
        self.dock_panes = dock_panes = [
            DockPane(id="tests.bogus_task.dock_pane_2", name="Dock Pane 2"),
            DockPane(id="tests.bogus_task.dock_pane_1", name="Dock Pane 1"),
        ]

        return dock_panes

    def _menu_bar_default(self):
        menu_bar = SMenuBar(
            SMenu(
                SGroup(
                    group_factory=DockPaneToggleGroup,
                    id="tests.bogus_task.DockPaneToggleGroup",
                ),
                id="View",
                name="&View",
            )
        )

        return menu_bar


class DockPaneToggleGroupTestCase(unittest.TestCase):
    @unittest.skipIf(USING_WX, "TaskWindowBackend is not implemented in WX")
    def setUp(self):
        self.gui = GUI()

        # Set up the bogus task with its window.
        self.task = BogusTask()

        self.window = window = TaskWindow()
        window.add_task(self.task)

        self.task_state = window._get_state(self.task)

        # Fish the dock pane toggle group from the menu bar manager.
        dock_pane_toggle_group = []

        def find_doc_pane_toggle(item):
            if item.id == "tests.bogus_task.DockPaneToggleGroup":
                dock_pane_toggle_group.append(item)

        self.task_state.menu_bar_manager.walk(find_doc_pane_toggle)

        self.dock_pane_toggle_group = dock_pane_toggle_group[0]

    def tearDown(self):
        del self.task
        del self.task_state
        del self.dock_pane_toggle_group

        if self.window.control is not None:
            self.window.destroy()
            self.gui.process_events()
        del self.window
        del self.gui

    def get_dock_pane_toggle_action_names(self):
        names = [
            action_item.action.name
            for action_item in self.dock_pane_toggle_group.items
        ]

        return names

    # Tests ----------------------------------------------------------------

    def test_group_content_at_startup(self):
        # Check that there are 2 dock panes in the group at the beginning.
        self.assertEqual(2, len(self.dock_pane_toggle_group.items))

        # Names are sorted by the group.
        names = self.get_dock_pane_toggle_action_names()
        expected_names = ["Dock Pane 1", "Dock Pane 2"]
        self.assertEqual(list(sorted(expected_names)), list(sorted(names)))

    def test_react_to_dock_pane_added(self):
        # Add a dock pane to the task.
        self.task_state.dock_panes.append(
            DockPane(id="tests.bogus_task.dock_pane_0", name="Dock Pane 0")
        )

        # Check that there are 3 dock panes in the group.
        self.assertEqual(3, len(self.dock_pane_toggle_group.items))

        # Names are sorted by the group.
        names = self.get_dock_pane_toggle_action_names()
        expected_names = ["Dock Pane 0", "Dock Pane 1", "Dock Pane 2"]
        self.assertEqual(list(sorted(expected_names)), list(sorted(names)))

    def test_react_to_dock_pane_removed(self):
        # Remove a dock pane from the task.
        self.task_state.dock_panes.remove(self.task.dock_panes[0])

        # Check that there is only 1 dock pane left in the group.
        self.assertEqual(1, len(self.dock_pane_toggle_group.items))

        names = self.get_dock_pane_toggle_action_names()
        expected_names = ["Dock Pane 1"]
        self.assertEqual(list(sorted(expected_names)), list(sorted(names)))



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from itertools import chain, combinations
from unittest import TestCase

from ..font_parser import (
    DECORATIONS,
    GENERIC_FAMILIES,
    NOISE,
    STRETCHES,
    STYLES,
    VARIANTS,
    WEIGHTS,
    simple_parser,
)


class TestSimpleParser(TestCase):

    def test_empty(self):
        properties = simple_parser("")
        self.assertEqual(
            properties,
            {
                'family': ["default"],
                'size': 12.0,
                'weight': "normal",
                'stretch': "normal",
                'style': "normal",
                'variants': set(),
                'decorations': set(),
            },
        )

    def test_typical(self):
        properties = simple_parser(
            "10 pt bold condensed italic underline Helvetica sans-serif")
        self.assertEqual(
            properties,
            {
                'family': ["helvetica", "sans-serif"],
                'size': 10.0,
                'weight': "bold",
                'stretch': "condensed",
                'style': "italic",
                'variants': set(),
                'decorations': {"underline"},
            },
        )

    def test_noise(self):
        for noise in NOISE:
            with self.subTest(noise=noise):
                properties = simple_parser(noise)
                self.assertEqual(
                    properties,
                    {
                        'family': ["default"],
                        'size': 12.0,
                        'weight': "normal",
                        'stretch': "normal",
                        'style': "normal",
                        'variants': set(),
                        'decorations': set(),
                    },
                )

    def test_generic_families(self):
        for family in GENERIC_FAMILIES:
            with self.subTest(family=family):
                properties = simple_parser(family)
                self.assertEqual(
                    properties,
                    {
                        'family': [family],
                        'size': 12.0,
                        'weight': "normal",
                        'stretch': "normal",
                        'style': "normal",
                        'variants': set(),
                        'decorations': set(),
                    },
                )

    def test_size(self):
        for size in [12, 24, 12.5]:
            with self.subTest(size=size):
                properties = simple_parser(str(size))
                self.assertEqual(
                    properties,
                    {
                        'family': ["default"],
                        'size': size,
                        'weight': "normal",
                        'stretch': "normal",
                        'style': "normal",
                        'variants': set(),
                        'decorations': set(),
                    },
                )

    def test_weight(self):
        for weight in WEIGHTS:
            with self.subTest(weight=weight):
                properties = simple_parser(weight)
                self.assertEqual(
                    properties,
                    {
                        'family': ["default"],
                        'size': 12.0,
                        'weight': weight,
                        'stretch': "normal",
                        'style': "normal",
                        'variants': set(),
                        'decorations': set(),
                    },
                )

    def test_stretch(self):
        for stretch in STRETCHES:
            with self.subTest(stretch=stretch):
                properties = simple_parser(stretch)
                self.assertEqual(
                    properties,
                    {
                        'family': ["default"],
                        'size': 12.0,
                        'weight': "normal",
                        'stretch': stretch,
                        'style': "normal",
                        'variants': set(),
                        'decorations': set(),
                    },
                )

    def test_style(self):
        for style in STYLES:
            with self.subTest(style=style):
                properties = simple_parser(style)
                self.assertEqual(
                    properties,
                    {
                        'family': ["default"],
                        'size': 12.0,
                        'weight': "normal",
                        'stretch': "normal",
                        'style': style,
                        'variants': set(),
                        'decorations': set(),
                    },
                )

    def test_variant(self):
        for variant in VARIANTS:
            with self.subTest(variant=variant):
                properties = simple_parser(variant)
                self.assertEqual(
                    properties,
                    {
                        'family': ["default"],
                        'size': 12.0,
                        'weight': "normal",
                        'stretch': "normal",
                        'style': "normal",
                        'variants': {variant},
                        'decorations': set(),
                    },
                )

    def test_decorations(self):
        # get powerset iterator of DECORATIONS
        all_decorations = chain.from_iterable(
            combinations(DECORATIONS, n)
            for n in range(len(DECORATIONS) + 1)
        )
        for decorations in all_decorations:
            with self.subTest(decorations=decorations):
                properties = simple_parser(" ".join(decorations))
                self.assertEqual(
                    properties,
                    {
                        'family': ["default"],
                        'size': 12.0,
                        'weight': "normal",
                        'stretch': "normal",
                        'style': "normal",
                        'variants': set(),
                        'decorations': set(decorations),
                    },
                )



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Font conversion utilities

This module provides facilities for converting between pyface Font objects
and Qt QFont objects, trying to keep as much similarity as possible between
them.
"""

from pyface.qt.QtGui import QFont


qt_family_to_generic_family = {
    QFont.StyleHint.AnyStyle: 'default',
    QFont.StyleHint.System: 'default',
    QFont.Decorative: 'fantasy',
    QFont.Serif: 'serif',
    QFont.StyleHint.Cursive: 'cursive',
    QFont.SansSerif: 'sans-serif',
    QFont.StyleHint.Monospace: 'monospace',
    QFont.TypeWriter: 'typewriter',
}
generic_family_to_qt_family = {
    'default': QFont.StyleHint.System,
    'fantasy': QFont.Decorative,
    'decorative': QFont.Decorative,
    'serif': QFont.Serif,
    'roman': QFont.Serif,
    'cursive': QFont.StyleHint.Cursive,
    'script': QFont.StyleHint.Cursive,
    'sans-serif': QFont.SansSerif,
    'swiss': QFont.SansSerif,
    'monospace': QFont.StyleHint.Monospace,
    'modern': QFont.StyleHint.Monospace,
    'typewriter': QFont.TypeWriter,
    'teletype': QFont.TypeWriter,
}

weight_to_qt_weight = {
    100: QFont.Weight.Thin,
    200: QFont.Weight.ExtraLight,
    300: QFont.Weight.Light,
    400: QFont.Weight.Normal,
    500: QFont.Weight.Medium,
    600: QFont.Weight.DemiBold,
    700: QFont.Weight.Bold,
    800: QFont.Weight.ExtraBold,
    900: QFont.Weight.Black,
    1000: 99,
}
qt_weight_to_weight = {
    QFont.Weight.Thin: 'thin',
    QFont.Weight.ExtraLight: 'extra-light',
    QFont.Weight.Light: 'light',
    QFont.Weight.Normal: 'normal',
    QFont.Weight.Medium: 'medium',
    QFont.Weight.DemiBold: 'demi-bold',
    QFont.Weight.Bold: 'bold',
    QFont.Weight.ExtraBold: 'extra-bold',
    QFont.Weight.Black: 'black',
    99: 'extra-heavy',
}

style_to_qt_style = {
    'normal': QFont.Style.StyleNormal,
    'oblique': QFont.Style.StyleOblique,
    'italic': QFont.Style.StyleItalic,
}
qt_style_to_style = {value: key for key, value in style_to_qt_style.items()}


def font_to_toolkit_font(font):
    """ Convert a Pyface font to a Qfont.

    Parameters
    ----------
    font : pyface.font.Font
        The Pyface font to convert.

    Returns
    -------
    qt_font : QFont
        The best matching Qt font.
    """
    qt_font = QFont()
    families = []
    default_family = None

    for family in font.family:
        if family not in generic_family_to_qt_family:
            families.append(family)
        elif default_family is None:
            default_family = family

    if families and hasattr(qt_font, 'setFamilies'):
        # Qt 5.13 and later
        qt_font.setFamilies(families)
    elif families:
        qt_font.setFamily(families[0])
        # Note: possibily could use substitutions here,
        # but not sure if global (which would be bad, so we don't)

    if default_family is not None:
        qt_font.setStyleHint(generic_family_to_qt_family[default_family])

    qt_font.setPointSizeF(font.size)
    qt_font.setWeight(weight_to_qt_weight[font.weight_])
    qt_font.setStretch(int(font.stretch))
    qt_font.setStyle(style_to_qt_style[font.style])
    qt_font.setUnderline('underline' in font.decorations)
    qt_font.setStrikeOut('strikethrough' in font.decorations)
    qt_font.setOverline('overline' in font.decorations)
    if 'small-caps' in font.variants:
        qt_font.setCapitalization(QFont.Capitalization.SmallCaps)
    return qt_font


def toolkit_font_to_properties(toolkit_font):
    """ Convert a QFont to a dictionary of font properties.

    Parameters
    ----------
    toolkit_font : QFont
        The Qt QFont to convert.

    Returns
    -------
    properties : dict
        Font properties suitable for use in creating a Pyface Font.
    """
    family = []

    if toolkit_font.family():
        family.append(toolkit_font.family())
    if hasattr(toolkit_font, 'families'):
        # Qt 5.13 and later
        family.extend(toolkit_font.families())
    family.append(qt_family_to_generic_family[toolkit_font.styleHint()])

    size = toolkit_font.pointSizeF()
    style = qt_style_to_style[toolkit_font.style()]
    weight = map_to_nearest(toolkit_font.weight(), qt_weight_to_weight)
    stretch = toolkit_font.stretch()
    if stretch == 0:
        # stretch 0 means any stretch is allowed, we default to no stretch
        stretch = 100.0
    variants = set()
    if toolkit_font.capitalization() == QFont.Capitalization.SmallCaps:
        variants.add('small-caps')
    decorations = set()
    if toolkit_font.underline():
        decorations.add('underline')
    if toolkit_font.strikeOut():
        decorations.add('strikethrough')
    if toolkit_font.overline():
        decorations.add('overline')

    return {
        'family': family,
        'size': size,
        'weight': weight,
        'stretch': stretch,
        'style': style,
        'variants': variants,
        'decorations': decorations,
    }


def map_to_nearest(target, mapping):
    """ Given mapping with keys from 0 and 99, return closest value.

    Parameters
    ----------
    target : int
        The value to map.
    mapping : dict
        A dictionary with integer keys ranging from 0 to 99.

    Returns
    -------
    value : Any
        The value corresponding to the nearest key.  In the case of a tie,
        the first value is returned.
    """
    if target in mapping:
        return mapping[target]

    distance = 100
    nearest = None
    for key in mapping:
        if abs(target - key) < distance:
            distance = abs(target - key)
            nearest = key
    return mapping[nearest]



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!from unittest import TestCase

import unittest

from pyface.toolkit import toolkit

is_wx = (toolkit.toolkit == "wx")


class TestModalDialogTester(unittest.TestCase):

    @unittest.skipIf(is_wx, "wx is not supported")
    def test_import(self):
        from pyface.util.modal_dialog_tester import ModalDialogTester
        self.assertNotEqual(ModalDialogTester.__name__, "Unimplemented")



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
API for the ``pyface.action.schema`` subpackage.

Schemas and Aliases
-------------------

- :class:`~.ActionSchema`
- :class:`~.GroupSchema`
- :class:`~.MenuBarSchema`
- :class:`~.MenuSchema`
- :attr:`~.SGroup`
- :attr:`~.SMenu`
- :attr:`~.SMenuBar`
- :attr:`~.SToolBar`
- :class:`~.ToolBarSchema`

Builder and Schema Additions
----------------------------

- :class:`~.ActionManagerBuilder`
- :class:`~.SchemaAddition`

"""

from .action_manager_builder import ActionManagerBuilder
from .schema import (
    ActionSchema,
    GroupSchema,
    MenuBarSchema,
    MenuSchema,
    SGroup,
    SMenu,
    SMenuBar,
    SToolBar,
    ToolBarSchema,
)
from .schema_addition import SchemaAddition



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The Qt-specific implementation of the text field class """

from pyface.qt.QtGui import QLabel, QPixmap

from traits.api import provides

from pyface.ui.qt.fields.field import Field
from pyface.fields.i_image_field import IImageField, MImageField


@provides(IImageField)
class ImageField(MImageField, Field):
    """ The Qt-specific implementation of the image field class
    """

    # ------------------------------------------------------------------------
    # IWidget interface
    # ------------------------------------------------------------------------

    def _create_control(self, parent):
        """ Create the toolkit-specific control that represents the widget. """
        control = QLabel(parent)
        return control

    # ------------------------------------------------------------------------
    # Private interface
    # ------------------------------------------------------------------------

    def _get_control_value(self):
        """ Toolkit specific method to get the control's value. """
        # XXX we should really have a ToolkitImage subclass of Image and
        # which would be the correct subclass to return.
        return self.control.pixmap()

    def _set_control_value(self, value):
        """ Toolkit specific method to set the control's value. """
        if value is None:
            self._toolkit_value = None
            self.control.setPixmap(QPixmap())
        else:
            self._toolkit_value = self.value.create_bitmap()
            self.control.setPixmap(self._toolkit_value)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Helper functions for working with images

This module provides helper functions for converting between numpy arrays
and Qt wx.Images, as well as between the various image types in a standardized
way.
"""

from enum import IntEnum

import wx


class ScaleMode(IntEnum):
    fast = wx.IMAGE_QUALITY_NORMAL
    smooth = wx.IMAGE_QUALITY_HIGH


class AspectRatio(IntEnum):
    ignore = 0
    keep_constrain = 1
    keep_expand = 2


def image_to_bitmap(image):
    """ Convert a wx.Image to a wx.Bitmap.

    Parameters
    ----------
    image : wx.Image
        The wx.Image to convert.

    Return
    ------
    bitmap : wx.Bitmap
        The corresponding wx.Bitmap.
    """
    return image.ConvertToBitmap()


def bitmap_to_image(bitmap):
    """ Convert a wx.Bitmap to a wx.Image.

    Parameters
    ----------
    bitmap : wx.Bitmap
        The wx.Bitmap to convert.

    Return
    ------
    image : wx.Image
        The corresponding wx.Image.
    """
    return bitmap.ConvertToImage()


def bitmap_to_icon(bitmap):
    """ Convert a wx.Bitmap to a wx.Icon.

    Parameters
    ----------
    bitmap : wx.Bitmap
        The wx.Bitmap to convert.

    Return
    ------
    icon : wx.Icon
        The corresponding wx.Icon.
    """
    return wx.Icon(bitmap)


def resize_image(image, size, aspect_ratio=AspectRatio.ignore,
                 mode=ScaleMode.fast):
    """ Resize a toolkit image to the given size. """
    image_size = image.GetSize()
    width, height = _get_size_for_aspect_ratio(image_size, size, aspect_ratio)
    return image.Scale(width, height, mode)


def resize_bitmap(bitmap, size, aspect_ratio=AspectRatio.ignore,
                  mode=ScaleMode.fast):
    """ Resize a toolkit bitmap to the given size. """
    image = bitmap_to_image(bitmap)
    image = resize_image(image, size, aspect_ratio, mode)
    return image_to_bitmap(image)


def image_to_array(image):
    """ Convert a wx.Image to a numpy array.

    This copies the data returned from wx.

    Parameters
    ----------
    image : wx.Image
        The wx.Image that we want to extract the values from.  The format must
        be either RGB32 or ARGB32.

    Return
    ------
    array : ndarray
        An N x M x 4 array of unsigned 8-bit ints as RGBA values.
    """
    import numpy as np

    width, height = image.GetSize()
    rgb_data = np.array(image.GetData(), dtype='uint8').reshape(width, height, 3)
    if image.HasAlpha():
        alpha = np.array(image.GetAlpha(), dtype='uint8').reshape(width, height)
    else:
        alpha = np.full((width, height), 0xff, dtype='uint8')
    array = np.empty(shape=(width, height, 4), dtype='uint8')
    array[:, :, :3] = rgb_data
    array[:, :, 3] = alpha
    return array


def array_to_image(array):
    """ Convert a numpy array to a wx.Image.

    This copies the data before passing it to wx.

    Parameters
    ----------
    array : ndarray
        An N x M x {3, 4} array of unsigned 8-bit ints.  The image
        format is assumed to be RGB or RGBA, based on the shape.

    Return
    ------
    image : wx.Image
        The wx.Image created from the data.
    """
    if array.ndim != 3:
        raise ValueError("Array must be either RGB or RGBA values.")

    height, width, channels = array.shape

    if channels == 3:
        image = wx.Image(width, height, array.tobytes())
    elif channels == 4:
        image = wx.Image(
            width,
            height,
            array[..., :3].tobytes(),
            array[..., 3].tobytes(),
        )
    else:
        raise ValueError("Array must be either RGB or RGBA values.")

    return image


def _get_size_for_aspect_ratio(image_size, size, aspect_ratio):
    width, height = size
    image_width, image_height = image_size
    if aspect_ratio != AspectRatio.ignore:
        if aspect_ratio == AspectRatio.keep_constrain:
            scale = min(width/image_width, height/image_height)
        else:
            scale = max(width/image_width, height/image_height)
        width = int(round(scale * image_width))
        height = int(round(scale * image_height))
    return (width, height)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


import unittest

from pyface.qt import QtCore, QtGui
from pyface.ui.qt.workbench.split_tab_widget import _DragableTabBar


class TestSplitTabWidget(unittest.TestCase):

    def test_mouseReleaseEvent(self):
        widget = _DragableTabBar(None, None)
        event = QtGui.QMouseEvent(
            QtCore.QEvent.Type.MouseButtonRelease,
            QtCore.QPointF(0.0, 0.0),
            QtCore.QPointF(0.0, 0.0),
            QtCore.Qt.MouseButton.RightButton,
            QtCore.Qt.RightButton,
            QtCore.Qt.NoModifier,
        )

        # smoke test: should do nothing
        widget.mouseReleaseEvent(event)

        widget.destroy()



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A panel that is split in two either horizontally or vertically. """

import warnings

from pyface.split_widget import SplitWidget
from pyface.widget import Widget


class SplitPanel(Widget, SplitWidget):
    """ A panel that is split in two either horizontally or vertically. """

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, parent=None, **traits):
        """ Creates a new panel. """

        create = traits.pop("create", None)

        # Base class constructor.
        super().__init__(parent=parent, **traits)

        if create:
            # Create the widget's toolkit-specific control.
            self.create()
            warnings.warn(
                "automatic widget creation is deprecated and will be removed "
                "in a future Pyface version, code should not pass the create "
                "parameter and should instead call create() explicitly",
                DeprecationWarning,
                stacklevel=2,
            )
        elif create is not None:
            warnings.warn(
                "setting create=False is no longer required",
                DeprecationWarning,
                stacklevel=2,
            )

    def _create_control(self, parent):
        """ Create the toolkit control """
        return self._create_splitter(parent)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The Wx-specific implementation of the text field class """


from traits.api import provides

from pyface.fields.i_field import IField, MField
from pyface.ui.wx.layout_widget import LayoutWidget
from pyface.ui.wx.util.alignment import (
    get_alignment_style, set_alignment_style
)


@provides(IField)
class Field(MField, LayoutWidget):
    """ The Wx-specific implementation of the field class

    This is an abstract class which is not meant to be instantiated.
    """

    # ------------------------------------------------------------------------
    # Private interface
    # ------------------------------------------------------------------------

    def _get_control_alignment(self):
        """ Toolkit specific method to get the control's read_only state. """
        return get_alignment_style(self.control.GetWindowStyle())

    def _set_control_alignment(self, alignment):
        """ Toolkit specific method to set the control's read_only state. """
        old_style = self.control.GetWindowStyle()
        new_style = set_alignment_style(alignment, old_style)
        self.control.SetWindowStyle(new_style)
        self.control.Refresh()



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" The base interface for an image. """
from traits.api import Interface


class IImage(Interface):
    """ The base interface for an image.

    This provides the interface specification that different types of image
    classes need to provide to be used by Pyface.
    """

    def create_image(self, size=None):
        """ Creates a toolkit-specific image for this image.

        An image is a toolkit datastructure that is optimized for I/O,
        pixel-level access and modification, such as a wx.Image or a
        QImage.

        Parameters
        ----------
        size : (int, int) or None
            The desired size as a width, height tuple, or None if wanting
            default image size.  This is a *preferred* size and concrete
            implementations may or may not return an image of the precise
            size requested (indeed, this may be ignored).

        Returns
        -------
        image : toolkit image
            The toolkit image corresponding to the image and the specified
            size.
        """

    def create_bitmap(self, size=None):
        """ Creates a toolkit-specific bitmap image for this image.

        A bitmap is a toolkit datastructure that is optimized for rendering
        to the screen, such as a wx.Bitmap or a QPixmap.

        Parameters
        ----------
        size : (int, int) or None
            The desired size as a width, height tuple, or None if wanting
            default image size.  This is a *preferred* size and concrete
            implementations may or may not return an image of the precise
            size requested (indeed, this may be ignored).

        Returns
        -------
        image : toolkit bitmap
            The toolkit bitmap corresponding to the image and the specified
            size.
        """

    def create_icon(self, size=None):
        """ Creates a toolkit-specific icon for this image.

        An icon is a toolkit datastructure that holds several different
        variants of an image (eg. selected, disabled, etc.), such as a wx.Icon
        or a QIcon.  Most toolkits can automatically create these from other
        image classes.

        Parameters
        ----------
        size : (int, int) or None
            The desired size as a width, height tuple, or None if wanting
            default icon size.  This is a *preferred* size and concrete
            implementations may or may not return an icon of the precise
            size requested (indeed, this may be ignored).

        Returns
        -------
        image : toolkit icon
            The toolkit image corresponding to the image and the specified
            size as an icon.
        """



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import logging
import unittest

# importlib.resources is new in Python 3.7, and importlib.resources.files is
# new in Python 3.9, so for Python < 3.9 we must rely on the 3rd party
# importlib_resources package.
try:
    from importlib.resources import as_file, files
except ImportError:
    from importlib_resources import as_file, files

from pyface.util._optional_dependencies import optional_import

Image = None

with optional_import(
        "PIL",
        msg="PILImage not available due to missing pillow.",
        logger=logging.getLogger(__name__)):

    from PIL import Image
    from ..pil_image import PILImage

image_source = files("pyface.tests") / "images" / "core.png"


@unittest.skipIf(Image is None, "Pillow not available")
class TestPILImage(unittest.TestCase):

    def test_create_image(self):
        with as_file(image_source) as image_path:
            pil_image = Image.open(image_path)
            image = PILImage(pil_image)
            toolkit_image = image.create_image()
            self.assertIsNotNone(toolkit_image)
            self.assertEqual(image.image, pil_image)

    def test_create_bitmap(self):
        with as_file(image_source) as image_path:
            pil_image = Image.open(image_path)
            image = PILImage(pil_image)
            bitmap = image.create_bitmap()
            self.assertIsNotNone(bitmap)

    def test_create_icon(self):
        with as_file(image_source) as image_path:
            pil_image = Image.open(image_path)
            image = PILImage(pil_image)
            icon = image.create_icon()
            self.assertIsNotNone(icon)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The image field interface. """

from traits.api import Any, HasTraits

from pyface.fields.i_field import IField
from pyface.ui_traits import Image


class IImageField(IField):
    """ The image field interface.

    This is for a field that edits a IImage value.
    """

    #: The current value of the image field
    value = Image()


class MImageField(HasTraits):
    """ Mixin class for ImageField implementations """

    #: The current value of the image field
    value = Image()

    #: The toolkit image to display
    _toolkit_value = Any()



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


from pyface.testing.layout_widget_mixin import LayoutWidgetMixin


class FieldMixin(LayoutWidgetMixin):
    """ Mixin which provides standard methods for all fields. """

    def test_text_field_alignment(self):
        self._create_widget_control()

        self.widget.alignment = 'right'
        self.gui.process_events()

        self.assertEqual(self.widget._get_control_alignment(), 'right')



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from pyface.qt import QtCore
from pyface.qt.QtGui import QApplication, QKeyEvent

from ..console_widget import ConsoleWidget


class TestConsoleWidget(unittest.TestCase):

    def setUp(self):
        # ensure QApplication is set up
        if not QApplication.instance():
            QApplication([])

    def test_format_as_columns(self):
        # regression test for Pyface #1146
        widget = ConsoleWidget()
        text = widget._format_as_columns(
            ["one", "two", "three"]
        )

        self.assertEqual(
            text,
            "one  two  three\n"
        )

    def test_right_key_press(self):
        # regression test for Pyface #1147
        widget = ConsoleWidget()
        event = QKeyEvent(
            QtCore.QEvent.Type.KeyPress,
            QtCore.Qt.Key.Key_Right,
            QtCore.Qt.KeyboardModifier.NoModifier,
        )

        intercepted = widget._event_filter_console_keypress(event)

        self.assertTrue(intercepted)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The implementation of a dialog that allows the user to select a color.
"""

from .constant import OK
from .toolkit import toolkit_object


ColorDialog = toolkit_object("color_dialog:ColorDialog")


def get_color(parent, color, show_alpha=False):
    """ Convenience function that displays a color dialog.

    Parameters
    ----------
    parent : toolkit control
        The parent toolkit control for the modal dialog.
    color : Color or color description
        The initial Color object, rgb(a) tuple or a string holding a valid
        color description.
    show_alpha : bool
        Whether or not to show alpha channel information.

    Returns
    -------
    color : Color or None
        The selected color, or None if the user made no selection.
    """
    dialog = ColorDialog(parent=parent, color=color, show_alpha=show_alpha)
    result = dialog.open()
    if result == OK:
        return dialog.color
    else:
        return None



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


GENERIC_FAMILIES = {
    'default', 'fantasy', 'decorative', 'serif', 'roman', 'cursive', 'script',
    'sans-serif', 'swiss', 'monospace', 'modern', 'typewriter', 'teletype'
}
WEIGHTS = {
    'thin', 'extra-light', 'light', 'regular', 'medium', 'demi-bold',
    'bold', 'extra-bold', 'heavy', 'extra-heavy'
}
STRETCHES = {
    'ultra-condensed', 'extra-condensed', 'condensed',
    'semi-condensed', 'semi-expanded', 'expanded', 'extra-expanded',
    'ultra-expanded'
}
STYLES = {'italic', 'oblique'}
VARIANTS = {'small-caps'}
DECORATIONS = {'underline', 'strikethrough', 'overline'}
NOISE = {'pt', 'point', 'px', 'family'}


class FontParseError(ValueError):
    """An exception raised when font parsing fails."""
    pass


def simple_parser(description):
    """An extremely simple font description parser.

    The parser is simple, and works by splitting the description on whitespace
    and examining each resulting token for understood terms:

    Size
        The first numeric term is treated as the font size.

    Weight
        The following weight terms are accepted: 'thin', 'extra-light',
        'light', 'regular', 'medium', 'demi-bold', 'bold', 'extra-bold',
        'heavy', 'extra-heavy'.

    Stretch
        The following stretch terms are accepted: 'ultra-condensed',
        'extra-condensed', 'condensed', 'semi-condensed', 'semi-expanded',
        'expanded', 'extra-expanded', 'ultra-expanded'.

    Style
        The following style terms are accepted: 'italic', 'oblique'.

    Variant
        The following variant terms are accepted: 'small-caps'.

    Decorations
        The following decoration terms are accepted: 'underline',
        'strikethrough', 'overline'.

    Generic Families
        The following generic family terms are accepted: 'default', 'fantasy',
        'decorative', 'serif', 'roman', 'cursive', 'script', 'sans-serif',
        'swiss', 'monospace', 'modern', 'typewriter', 'teletype'.

    In addtion, the parser ignores the terms 'pt', 'point', 'px', and 'family'.
    Any remaining terms are combined into the typeface name.  There is no
    expected order to the terms.

    This parser is roughly compatible with the various ad-hoc parsers in
    TraitsUI and Kiva, allowing for the slight differences between them and
    adding support for additional options supported by Pyface fonts, such as
    stretch and variants.

    Parameters
    ----------
    description : str
        The font description to be parsed.

    Returns
    -------
    properties : dict
        Font properties suitable for use in creating a Pyface Font.

    Notes
    -----
    This is not a particularly good parser, as it will fail to properly
    parse something like "10 pt times new roman" or "14 pt computer modern"
    since they have generic font names as part of the font face name.
    """
    face = []
    generic_family = ""
    size = None
    weight = "normal"
    stretch = "normal"
    style = "normal"
    variants = set()
    decorations = set()
    for word in description.lower().split():
        if word in NOISE:
            continue
        elif word in GENERIC_FAMILIES:
            generic_family = word
        elif word in WEIGHTS:
            weight = word
        elif word in STRETCHES:
            stretch = word
        elif word in STYLES:
            style = word
        elif word in VARIANTS:
            variants.add(word)
        elif word in DECORATIONS:
            decorations.add(word)
        else:
            if size is None:
                try:
                    size = float(word)
                    continue
                except ValueError:
                    pass
            face.append(word)

    family = []
    if face:
        family.append(" ".join(face))
    if generic_family:
        family.append(generic_family)
    if not family:
        family = ["default"]
    if size is None:
        size = 12

    return {
        'family': family,
        'size': size,
        'weight': weight,
        'stretch': stretch,
        'style': style,
        'variants': variants,
        'decorations': decorations,
    }



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" The default resource manager.

A resource manager locates and loads application resources such as images and
sounds etc.
"""

import collections.abc
import glob
import inspect
import os
from os.path import join
import types
from zipfile import is_zipfile, ZipFile

# importlib.resources is new in Python 3.7, and importlib.resources.files is
# new in Python 3.9, so for Python < 3.9 we must rely on the 3rd party
# importlib_resources package.
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

from traits.api import HasTraits, Instance, List
from traits.util.resource import get_path

from pyface.resource.resource_factory import ResourceFactory
from pyface.resource.resource_reference import ImageReference


class ResourceManager(HasTraits):
    """ The default resource manager.

    A resource manager locates and loads application resources such as images
    and sounds etc.
    """

    # Allowed extensions for image resources.
    IMAGE_EXTENSIONS = [".png", ".jpg", ".bmp", ".gif", ".ico"]

    # A list of additional search paths. These paths are fallbacks, and hence
    # have lower priority than the paths provided by resource objects.
    extra_paths = List()

    # The resource factory is responsible for actually creating resources.
    # This is used so that (for example) different GUI toolkits can create
    # a images in the format that they require.
    resource_factory = Instance(ResourceFactory)

    # ------------------------------------------------------------------------
    # 'ResourceManager' interface.
    # ------------------------------------------------------------------------

    def locate_image(self, image_name, path, size=None):
        """ Locates an image.

        Parameters
        ----------
        image_name : str
            Name of the image file.
        path : list of (str or ModuleType)
            Paths from which image files will be searched. Note that for each
            path, a subdirectory named 'images' will be search first.
            The first match will be returned.
        size : tuple of (m: int, n: int), optional
            Specific size of the image requested. If provided, then
            the subdirectory ``images/{m}x{n}`` will be searched first,
            followed by the ``images`` subdirectory and its containing folder.
            Default is None.

        Returns
        -------
        image_ref : ImageReference or None
            ImageReference to the image found, or None if no matching images
            are found.
        """

        if not isinstance(path, collections.abc.Sequence):
            path = [path]

        resource_path = []
        for item in list(path) + self.extra_paths:
            if isinstance(item, str):
                resource_path.append(item)
            elif isinstance(item, types.ModuleType):
                resource_path.append(item)
            else:
                resource_path.extend(self._get_resource_path(item))

        return self._locate_image(image_name, resource_path, size)

    def load_image(self, image_name, path, size=None):
        """ Loads an image. """

        reference = self.locate_image(image_name, path, size)
        if reference is not None:
            image = reference.load()

        else:
            image = None

        return image

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _locate_image(self, image_name, resource_path, size):
        """ Attempts to locate an image resource.

        If the image is found, an image resource reference is returned.
        If the image is NOT found None is returned.

        """

        # If the image name contains a file extension (eg. '.jpg') then we will
        # only accept an an EXACT filename match.
        basename, extension = os.path.splitext(image_name)
        if len(extension) > 0:
            extensions = [extension]
            pattern = image_name

        # Otherwise, we will search for common image suffixes.
        else:
            extensions = self.IMAGE_EXTENSIONS
            pattern = image_name + ".*"

        # Try the 'images' sub-directory first (since that is commonly
        # where we put them!).  If the image is not found there then look
        # in the directory itself.
        if size is None:
            subdirs = ["images", ""]
        else:
            subdirs = ["images/%dx%d" % (size[0], size[1]), "images", ""]

        # Concrete image filenames to be searched
        image_filenames = [basename + extension for extension in extensions]

        for dirname in resource_path:

            # If we come across a reference to a module, try and find the
            # image inside of an .egg, .zip, etc.
            if isinstance(dirname, types.ModuleType):
                try:
                    data = _find_resource_data(
                        dirname, subdirs, image_filenames
                    )
                except OSError:
                    continue
                else:
                    return ImageReference(
                        self.resource_factory, data=data
                    )

            # Is there anything resembling the image name in the directory?
            for path in subdirs:
                filenames = glob.glob(join(dirname, path, pattern))
                for filename in filenames:
                    not_used, extension = os.path.splitext(filename)
                    if extension in extensions:
                        reference = ImageReference(
                            self.resource_factory, filename=filename
                        )

                        return reference

            # Is there an 'images' zip file in the directory?
            zip_filename = join(dirname, "images.zip")
            if os.path.isfile(zip_filename):
                zip_file = ZipFile(zip_filename, "r")
                # Try the image name itself, and then the image name with
                # common images suffixes.
                for extension in extensions:
                    try:
                        image_data = zip_file.read(basename + extension)
                        reference = ImageReference(
                            self.resource_factory, data=image_data
                        )

                        return reference

                    except:
                        pass

            # is this a path within a zip file?
            # first, find the zip file in the path
            filepath = dirname
            zippath = ""
            while (
                not is_zipfile(filepath)
                and os.path.splitdrive(filepath)[1].startswith("\\")
                and os.path.splitdrive(filepath)[1].startswith("/")
            ):
                filepath, tail = os.path.split(filepath)
                if zippath != "":
                    zippath = tail + "/" + zippath
                else:
                    zippath = tail

            # if we found a zipfile, then look inside it for the image!
            if is_zipfile(filepath):

                zip_file = ZipFile(filepath)
                for subpath in ["images", ""]:
                    for extension in extensions:
                        try:
                            # this is a little messy. since zip files don't
                            # recognize a leading slash, we have to be very
                            # particular about how we build this path when
                            # there are empty strings
                            if zippath != "":
                                path = zippath + "/"
                            else:
                                path = ""

                            if subpath != "":
                                path = path + subpath + "/"

                            path = path + basename + extension
                            # now that we have the path we can attempt to load
                            # the image
                            image_data = zip_file.read(path)
                            reference = ImageReference(
                                self.resource_factory, data=image_data
                            )

                            # if there was no exception then return the result
                            return reference

                        except:
                            pass

        return None

    def _get_resource_path(self, object):
        """ Returns the resource path for an object. """

        if hasattr(object, "resource_path"):
            resource_path = object.resource_path

        else:
            resource_path = self._get_default_resource_path(object)

        return resource_path

    def _get_default_resource_path(self, object):
        """ Returns the default resource path for an object. """

        resource_path = []
        for klass in inspect.getmro(object.__class__):
            try:
                resource_path.append(get_path(klass))

            # We get an attribute error when we get to a C extension type (in
            # our case it will most likley be 'CHasTraits'.  We simply ignore
            # everything after this point!
            except AttributeError:
                break

        return resource_path


def _get_package_data(module, rel_path):
    """ Return package data in bytes for the given module and resource path.

    Parameters
    ----------
    module : ModuleType
        A module from which package data will be discovered.
        If the module name does not conform to the package requirement, then
        its "__file__" attribute is used for locating the directory to search
        for resource files.
    rel_path : str
        "/"-separated path for loading data file.

    Returns
    -------
    data : bytes
        Loaded data in bytes.

    Raises
    ------
    OSError
        If the path referenced does not resolve to an existing file or the
        file cannot be read.
    """

    if (module.__spec__ is None
            or module.__spec__.submodule_search_locations is None):
        module_dir_path = os.path.dirname(module.__file__)
        path = os.path.join(module_dir_path, *rel_path.split("/"))
        with open(path, "rb") as fp:
            return fp.read()

    return (
        files(module).joinpath(rel_path).read_bytes()
    )


def _find_resource_data(module, subdirs, filenames):
    """ For the given module, search directories and names, find a matching
    resource file and return its content as bytes.

    Parameters
    ----------
    module : ModuleType
        A module from which package data will be discovered.
        If the module name does not conform to the package requirement, then
        its "__file__" attribute is used for locating the directory to search
        for resource files.
    subdirs : list of str
        Name of subdirectories to try. Each value can be a "/"-separated
        string to represent more nested subdirectories.
    filenames : list of str
        File names to try.

    Returns
    -------
    data : bytes
        Loaded data in bytes.

    Raises
    ------
    OSError
        If the path referenced does not resolve to an existing file or the
        file cannot be read.
    """
    for path in subdirs:
        for filename in filenames:
            searchpath = "%s/%s" % (path, filename)
            try:
                return _get_package_data(module, searchpath)
            except OSError:
                pass
    raise OSError(
        "Unable to load data for the given module and search paths."
    )



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Test for pyface.api """

import unittest


class TestApi(unittest.TestCase):
    """ Test importable items in any environment."""

    def test_api_importable(self):
        # make sure api is importable with the most minimal
        # required dependencies, including in the absence of toolkit backends.
        from pyface.tasks import api   # noqa: F401

    def test_public_attrs(self):
        # make sure everything advertised by dir() is available except optional
        from pyface.tasks import api

        attrs = [
            name
            for name in dir(api)
            if not name.startswith('_')
        ]
        for attr in attrs:
            with self.subTest(attr=attr):
                self.assertIsNotNone(getattr(api, attr, None))



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


import contextlib
import threading
import warnings

from pyface.i_gui import IGUI
from pyface.qt import QtCore, QtGui
from traits.api import HasStrictTraits, Instance


class ConditionTimeoutError(RuntimeError):
    pass


@contextlib.contextmanager
def dont_quit_when_last_window_closed(qt_app):
    """
    Suppress exit of the application when the last window is closed.

    """
    flag = qt_app.quitOnLastWindowClosed()
    qt_app.setQuitOnLastWindowClosed(False)
    try:
        yield
    finally:
        qt_app.setQuitOnLastWindowClosed(flag)


class EventLoopHelper(HasStrictTraits):

    qt_app = Instance(QtGui.QApplication)

    gui = Instance(IGUI)

    def event_loop_with_timeout(self, repeat=2, timeout=10.0):
        """Helper function to send all posted events to the event queue and
        wait for them to be processed.  This runs the real event loop and
        does not emulate it with QApplication.processEvents.

        Parameters
        ----------
        repeat : int
            Number of times to process events. Default is 2.
        timeout: float, optional, keyword only
            Number of seconds to run the event loop in the case that the trait
            change does not occur. Default value is 10.0.

        Notes
        -----
            `timeout` is rounded to the nearest millisecond.
        """

        def repeat_loop(condition, repeat):
            # We sendPostedEvents to ensure that enaml events are processed
            self.qt_app.sendPostedEvents()
            repeat = repeat - 1
            if repeat <= 0:
                self.gui.invoke_later(condition.set)
            else:
                self.gui.invoke_later(
                    repeat_loop, condition=condition, repeat=repeat
                )

        condition = threading.Event()
        self.gui.invoke_later(repeat_loop, repeat=repeat, condition=condition)
        self.event_loop_until_condition(
            condition=condition.is_set, timeout=timeout
        )

    def event_loop(self, repeat=1):
        """Emulates an event loop `repeat` times with
        QApplication.processEvents.

        Parameters
        ----------
        repeat : int
            Number of times to process events. Default is 1.
        """
        for i in range(repeat):
            self.qt_app.sendPostedEvents()
            self.qt_app.processEvents()

    def event_loop_until_condition(self, condition, timeout=10.0):
        """Runs the real Qt event loop until the provided condition evaluates
        to True.

        Notes
        -----

        This runs the real Qt event loop, polling the condition every 50 ms and
        returning as soon as the condition becomes true. If the condition does
        not become true within the given timeout, a ConditionTimeoutError is
        raised.

        Because the state of the condition is only polled every 50 ms, it
        may fail to detect transient states that appear and disappear within
        a 50 ms window.  Code should ensure that any state that is being
        tested by the condition cannot revert to a False value once it becomes
        True.

        Parameters
        ----------
        condition : Callable
            A callable to determine if the stop criteria have been met. This
            should accept no arguments.

        timeout : float
            Number of seconds to run the event loop in the case that the trait
            change does not occur.

            `timeout` is rounded to the nearest millisecond.

        Raises
        ------

        Raises ConditionTimeoutError if the timeout occurs before the condition
        is satisfied.  If the event loop exits before the condition evaluates
        to True or times out, a RuntimeWarning will be generated.

        In either of these cases, the message will indicate whether the
        condition was ever successfully evaluated (which may indicate an error
        in the condition's code) or whether it always evalutated to False.
        """

        condition_result = None
        timed_out = False

        def handler():
            nonlocal condition_result

            condition_result = bool(condition_result or condition())
            if condition_result:
                self.qt_app.exit()

        def do_timeout():
            nonlocal timed_out

            timed_out = True
            self.qt_app.exit()

        # Make sure we don't get a premature exit from the event loop.
        with dont_quit_when_last_window_closed(self.qt_app):
            condition_timer = QtCore.QTimer()
            condition_timer.setInterval(50)
            condition_timer.timeout.connect(handler)
            timeout_timer = QtCore.QTimer()
            timeout_timer.setSingleShot(True)
            timeout_timer.setInterval(round(timeout * 1000))
            timeout_timer.timeout.connect(do_timeout)
            timeout_timer.start()
            condition_timer.start()
            try:
                if hasattr(self.qt_app, 'exec'):
                    self.qt_app.exec()
                else:
                    self.qt_app.exec_()
                if not condition_result:
                    if condition_result is None:
                        status = "without evaluating condition"
                    else:
                        status = "without condition evaluating to True"
                    if timed_out:
                        raise ConditionTimeoutError(f"Timed out {status}.")
                    else:
                        warnings.warn(
                            RuntimeWarning(
                                f"Event loop exited early {status}."
                            )
                        )
            finally:
                timeout_timer.stop()
                condition_timer.stop()

    @contextlib.contextmanager
    def delete_widget(self, widget, timeout=1.0):
        """Runs the real Qt event loop until the widget provided has been
        deleted.  Raises ConditionTimeoutError on timeout.

        Parameters
        ----------
        widget : QObject
            The widget whose deletion will stop the event loop.

        timeout : float
            Number of seconds to run the event loop in the case that the
            widget is not deleted.

        Notes
        -----
            `timeout` is rounded to the nearest millisecond.
        """
        timer = QtCore.QTimer()
        timer.setSingleShot(True)
        timer.setInterval(round(timeout * 1000))
        timer.timeout.connect(self.qt_app.quit)
        widget.destroyed.connect(self.qt_app.quit)
        yield
        timer.start()
        if hasattr(self.qt_app, 'exec'):
            self.qt_app.exec()
        else:
            self.qt_app.exec_()
        if not timer.isActive():
            # We exited the event loop on timeout
            raise ConditionTimeoutError(
                "Could not destroy widget before timeout: {!r}".format(widget)
            )
